import json
import pathlib
import re
import typing as t


class MissingAllowedIDS:
    pass


MISSING_ALLOWED_IDS = MissingAllowedIDS()


class IDSNotAllowed(Exception):
    pass


class IDSInvalidWrite(Exception):
    pass


class IDSValidateWrongParams(Exception):
    pass


class NoAllowedIdsSpecified(Exception):
    pass


class InvalidIDSString(ValueError):
    pass


class IDS:
    """Abstraction for IDS Name"""

    def __init__(self, namespace: str, slug: str, version: str):
        self.namespace = namespace
        self.slug = slug
        self.version = version

    @classmethod
    def from_dict(cls, ids_dict: dict):
        return cls(ids_dict["namespace"], ids_dict["slug"], ids_dict["version"])

    @classmethod
    def from_str(cls, ids_str: str):
        regex = re.compile(r"(.*)/(.*):(.*)")
        match_result = regex.match(ids_str)
        if match_result is None:
            raise InvalidIDSString(
                f"Received invalid IDS string: '{ids_str}'. IDS strings are expected to have format "
                f"<namespace>/<slug>:<version>."
            )
        namespace, slug, version = match_result.groups()
        return IDS(namespace=namespace, slug=slug, version=version)

    def to_tuple(self):
        return self.namespace, self.slug, self.version

    def __str__(self):
        return f"{self.namespace}/{self.slug}:{self.version}"

    def __repr__(self):
        return (
            f"IDS(namespace={self.namespace}, slug={self.slug}, version={self.version})"
        )

    def __eq__(self, item):
        return str(item) == str(self)

    def __hash__(self):
        return hash(str(self))

    def __bool__(self) -> bool:
        if self.namespace and self.slug and self.version:
            return True
        else:
            return False


class AllowedIDS:
    """Abstraction for 'allowedIds' JSON object"""

    allowed_ids: t.Iterable[IDS]

    def __init__(self, allowed_ids: t.Optional[t.Union[IDS, t.Iterable[IDS]]]):
        if isinstance(allowed_ids, IDS):
            self.allowed_ids = [allowed_ids]
        else:
            self.allowed_ids = allowed_ids

    @classmethod
    def from_allowedIds(cls, allowed_ids: t.Union[dict, t.List[dict], None]):
        """
        Read config.json to get the value of 'allowedIds'
        if 'allowedIds' is null, return None
        else return list of IDS object
        """

        if isinstance(allowed_ids, dict):
            return cls(IDS.from_dict(allowed_ids))
        elif isinstance(allowed_ids, list):
            allowed_ids = [IDS.from_dict(ids) for ids in allowed_ids]
            allowed_ids = list(set(allowed_ids))
            return cls(allowed_ids)
        elif allowed_ids is None:
            return cls(allowed_ids)
        else:
            raise ValueError(
                "'allowedIds' for function config (in config.json) must be one of the following: \n"
                "1. JSON object representing IDS\n"
                "2. List of JSON objects, each representing IDS\n"
                "3. null\n"
            )

    def is_single(self):
        return len(self.allowed_ids) == 1

    def get_first(self):
        return self.allowed_ids[0]

    def is_reconcilable(self, ids_requested_to_be_written: t.Optional[str]) -> bool:
        if not self.allowed_ids:
            raise IDSInvalidWrite(
                "Task Script is not allowed to write or validate IDS.\n"
                "Please make sure 'allowedIds' for function config (in config.json) is one of the following: \n"
                "1. JSON object representing IDS\n"
                "2. List of JSON objects, each representing IDS\n"
            )

        if ids_requested_to_be_written:
            ids_obj = IDS.from_str(ids_requested_to_be_written)
            if ids_obj not in self.allowed_ids:
                raise IDSNotAllowed(
                    f"Task Script is trying to write IDS for {ids_obj}."
                    f"However, only following IDS are allowed {self.allowed_ids}"
                )
        else:
            if len(self.allowed_ids) > 1:
                raise IDSInvalidWrite(
                    f"Could not determine which IDS to use. \n"
                    f"'ids_requested_to_be_written' is set to None."
                    f"And multiple IDS are allowed : {self.allowed_ids}"
                )
        return True

    def get_ids_to_be_written(
        self, ids_requested_to_be_written: t.Optional[str]
    ) -> t.Optional[str]:
        """
        'ids_requested_to_be_written' is safe to use if it is present in 'allowedIds'.
        If 'ids_requested_to_be_written' is None, then safely determine which IDS should the
        task-script write.
        """
        if self.is_reconcilable(ids_requested_to_be_written):
            if not ids_requested_to_be_written:
                return (
                    str(self.allowed_ids[0])
                    if isinstance(self.allowed_ids, list)
                    else str(self.allowed_ids)
                )
            return ids_requested_to_be_written

    def __eq__(self, other) -> bool:
        if isinstance(other, MissingAllowedIDS):
            return False
        return str(self.allowed_ids) == str(other.allowed_ids)


class FunctionConfig:
    """Abstraction for function configuration in config.json"""

    def __init__(self, config_dict: dict = {}):
        self._config_dict = config_dict
        self.allowed_ids = self._get_allowed_ids()

    def _get_allowed_ids(self) -> t.Union[AllowedIDS, MissingAllowedIDS]:
        if "allowedIds" not in self._config_dict:
            return MISSING_ALLOWED_IDS
        else:
            allowed_ids = self._config_dict["allowedIds"]
            return AllowedIDS.from_allowedIds(allowed_ids)

    @classmethod
    def from_task_script_config(
        cls, task_script_config: dict, func_slug: str
    ) -> "FunctionConfig":
        """
        Given task_script_config and func_slug, return FunctionConfig object
        """
        if not task_script_config:
            return cls({})

        functions = task_script_config.get("functions", [])
        fx_config = [fx for fx in functions if fx.get("slug") == func_slug]
        fx_config = fx_config[0] if fx_config else {}
        return cls(fx_config)

    @classmethod
    def from_task_script_dir(cls, func_dir: str, func_slug: str) -> "FunctionConfig":
        """Read function's configuration from task-script's config.json

        Args:
            func_dir (str): path of the directory containing task-script
            func_slug (str): Slug for the function being called

        Returns:
            FunctionConfig object
        """
        config_file = pathlib.Path(func_dir) / "config.json"

        if not config_file.exists():
            return cls.from_task_script_config({}, func_slug)

        with open(config_file, "r") as config:
            config_dict = json.load(config)
            return cls.from_task_script_config(config_dict, func_slug)



from enum import Enum
from typing import Union


class ObjectType(Enum):
    FILE = "file"
    _DEBUG = "_debug"


short_names = {ObjectType.FILE: "fi"}


def remove_trailing_slash(url: str) -> str:
    return url.rstrip("/")


def to_object_type(value: str) -> ObjectType:
    object_type = next(
        (object_type for object_type in ObjectType if object_type.value == value), None
    )
    if object_type is None:
        raise ValueError(f"{value} is not a valid object type")
    else:
        return object_type


def ensure_object_type(value: Union[ObjectType, str]) -> ObjectType:
    if isinstance(value, ObjectType):
        return value
    elif isinstance(value, str):
        return to_object_type(value)
    else:
        raise ValueError("Object type must be either an ObjectType or a string")


def select_short_name(object_type: ObjectType) -> str:
    if object_type in short_names:
        return short_names[object_type]
    else:
        raise NotImplementedError(
            f"Permalink cannot be created for object of type {object_type.value}"
        )


def get_permalink(
    platform_api_url: str, object_type: Union[ObjectType, str], uuid: str
):
    if not isinstance(uuid, str):
        raise ValueError("UUID must be a string")
    if not isinstance(platform_api_url, str):
        raise ValueError("Platform API URL must be a string")
    api_url = remove_trailing_slash(platform_api_url)
    surely_object_type = ensure_object_type(object_type)
    short_name = select_short_name(surely_object_type)
    return f"{api_url}/o/{short_name}/{uuid}"



"""Datalake functions"""

import gzip
import io
import os
import re
import shutil
import tempfile
from pathlib import Path
from typing import Collection, Dict, Optional, Tuple, Union
from urllib.parse import parse_qs, urlencode
from uuid import uuid4

import boto3
import simplejson as json
import smart_open
from botocore.client import Config

from .__util_metadata import FIELDS
from .__util_versioned_ref import VersionedRef
from .data_model import AnyLabel, IdsUtilDict
from .encoders import DataclassEncoder

WRITE_ALLOWED_CATEGORIES = ["IDS", "PROCESSED", "TMP"]
DISABLE_GZIP = bool(os.environ.get("DISABLE_GZIP"))
ENV = os.environ.get("ENV")
AWS_REGION = os.environ.get("AWS_REGION")
# don't use multipart if file size is smaller than 100MB
MULTIPART_SIZE = int(os.environ.get("MULTIPART_SIZE_MB", 100)) * 2**20
LARGE_FILE_SIZE_THRESHOLD_BYTES = (
    5 * 2**30
)  # We have to handle files that are > 5 GB differently


def lowerMetadataKeys(metadata: dict) -> dict:
    """Lowercases all keys in a dict"""
    return {k.lower(): v for k, v in metadata.items()}


def resolveCustomMetadataAndTags(metadata: dict) -> dict:
    custom_metadata_str = metadata.get(FIELDS["CUSTOM_METADATA"], "") or ""
    custom_tags_str = metadata.get(FIELDS["CUSTOM_TAGS"], "") or ""
    if not custom_tags_str:
        custom_tags = []
    else:
        custom_tags = custom_tags_str.split(",")

    return {
        "custom_metadata": parse_query(custom_metadata_str),
        "custom_tags": custom_tags,
    }


def getOrEmptyString(dic, key, default=""):
    val = dic.get(key, default)
    if val is None:
        return default
    return val


class InvalidPathException(Exception):
    """Exception for invalid path, listing path and reason"""

    def __init__(self, path, reason):
        super().__init__(f"Invalid path {path}: {reason}")
        self.path = path
        self.reason = reason


def getUpdatedPipelineHistoryStr(pipeline_id, existing_pipeline_history):
    new_pipeline_history = ""
    if existing_pipeline_history == "":
        new_pipeline_history = pipeline_id
    else:
        if pipeline_id in existing_pipeline_history:
            new_pipeline_history = existing_pipeline_history
        else:
            new_pipeline_history = existing_pipeline_history + "," + pipeline_id
    return new_pipeline_history


def parse_query(query_string: str) -> dict:
    result = dict()
    if not isinstance(query_string, str):
        return result
    if "?" in query_string:
        query_string = query_string.split("?")[-1]
    if "#" in query_string:
        query_string = query_string.split("#")[0]
    for key, values in parse_qs(query_string).items():
        if len(values) == 1:
            result[key] = values[0]
        else:
            result[key] = values
    return result


class S3FileobjUploader:
    """AWS S3 File object uploader"""

    def __init__(
        self, s3, fileobj, params, options={"disable_gzip": False, "compresslevel": 5}
    ):
        self.s3 = s3
        self.fileobj = fileobj
        self.params = params
        self.disable_gzip = options.get("disable_gzip", False)
        self.compresslevel = options.get("compresslevel", 5)
        self.stream = io.BytesIO()
        self.part_count = 0
        self.multipart = None
        self.parts = []
        if not self.disable_gzip:
            self.compressor = gzip.GzipFile(
                fileobj=self.stream, mode="wb", compresslevel=self.compresslevel
            )

    def _upload_part(self):
        print(f"upload multipart {self.part_count}")
        if self.part_count == 0:
            self.multipart = self.s3.create_multipart_upload(**self.params)
        self.part_count += 1
        self.stream.seek(0)
        part = self.s3.upload_part(
            Body=self.stream,
            Bucket=self.multipart["Bucket"],
            Key=self.multipart["Key"],
            PartNumber=self.part_count,
            UploadId=self.multipart["UploadId"],
        )
        self.parts.append({**part, "PartNumber": self.part_count})
        self.stream.seek(0)
        self.stream.truncate()

    def _upload_last_part(self):
        if self.part_count == 0:
            self.stream.seek(0)
            return self.s3.put_object(Body=self.stream, **self.params)
        self._upload_part()
        parts = []
        for part in self.parts:
            parts.append({"ETag": part["ETag"], "PartNumber": part["PartNumber"]})
        return self.s3.complete_multipart_upload(
            Bucket=self.multipart["Bucket"],
            Key=self.multipart["Key"],
            UploadId=self.multipart["UploadId"],
            MultipartUpload={"Parts": parts},
        )

    def upload(self):
        while True:
            chunk = self.fileobj.read(1024 * 1024)
            if not chunk:
                if not self.disable_gzip:
                    self.compressor.close()
                return self._upload_last_part()
            if self.disable_gzip:
                self.stream.write(chunk)
            else:
                self.compressor.write(chunk)
            if self.stream.tell() >= MULTIPART_SIZE:
                self._upload_part()


class MissingValidatorException(Exception):
    """An IDS cannot be validated because validator could not be found"""


class InvalidFileCategoryError(Exception):
    """The 'file category' is not valid. See :const:`~WRITE_ALLOWED_CATEGORIES`."""


class Datalake:
    """S3 Datalake class"""

    def __init__(self, endpoint, ids_util: Optional[IdsUtilDict] = None):
        self._ids_util = ids_util
        if endpoint:
            self.s3 = boto3.client(
                "s3",
                endpoint_url=endpoint,
                aws_access_key_id="123",
                aws_secret_access_key="abc",
                region_name=AWS_REGION,
                config=Config(signature_version="s3v4"),
            )
        else:
            self.s3 = boto3.client(
                "s3", region_name=AWS_REGION, config=Config(signature_version="s3v4")
            )

    def get_s3_head(self, file: dict):
        """Retrieve first object from S3 bucket"""
        bucket = file["bucket"]
        file_key = file["fileKey"]
        if "version" in file:
            file_version = file["version"]
            head = self.s3.head_object(
                Bucket=bucket, Key=file_key, VersionId=file_version
            )
        else:
            head = self.s3.head_object(Bucket=bucket, Key=file_key)

        return head

    def get_file_meta(self, file):
        """Return lowercased metadata dictionary of file argument"""
        head = self.get_s3_head(file)
        return lowerMetadataKeys(head.get("Metadata"))

    def read_file(self, file, form="body", tmp_dir="/tmp"):
        """Read file from S3"""
        bucket = file["bucket"]
        file_key = file["fileKey"]
        if "version" in file:
            kwargs = {"VersionId": file["version"]}
        else:
            kwargs = {}

        if form == "body":
            response = self.s3.get_object(Bucket=bucket, Key=file_key, **kwargs)
        elif form in ["file_obj", "download"]:
            response = self.s3.head_object(Bucket=bucket, Key=file_key, **kwargs)
        else:
            raise ValueError(
                f"Invalid form={form}; supported values are body, file_obj and download"
            )

        status_code = response.get("ResponseMetadata", {}).get("HTTPStatusCode")
        if status_code != 200:
            print({"level": "error", "message": response})
            raise Exception("Invalid response code")

        meta = lowerMetadataKeys(response.get("Metadata"))
        result = {"metadata": meta, **resolveCustomMetadataAndTags(meta)}

        if form == "body":
            content = response.get("Body").read()
            if response.get("ContentEncoding") == "gzip":
                result["body"] = gzip.decompress(content)
            else:
                result["body"] = content
        else:
            file_obj = smart_open.open(
                f"s3://{bucket}/{file_key}",
                "rb",
                transport_params={
                    "version_id": file.get("version"),
                    "client": self.s3,
                },
            )
            if response.get("ContentEncoding") == "gzip":
                file_obj = gzip.open(file_obj)

            if form == "download":
                with tempfile.NamedTemporaryFile(
                    mode="wb", delete=False, dir=tmp_dir
                ) as temp:
                    shutil.copyfileobj(file_obj, temp)
                    result["download"] = temp.name
                file_obj.close()
            else:
                result["file_obj"] = file_obj

        return result

    @classmethod
    def _is_file_path_valid(cls, filepath) -> Tuple[bool, str]:
        filepath = Path(filepath)
        if filepath.name == ".." or ".." in (
            parent.name for parent in filepath.parents
        ):
            return False, 'Path cannot contain a directory ".."'
        return True, ""

    @classmethod
    def _should_disable_gzip(cls, compression_level: int) -> bool:
        if compression_level == 0:
            return True
        else:
            return DISABLE_GZIP

    def write_file(
        self,
        context,
        content: Union[str, bytes, io.BytesIO, Dict],
        file_name: str,
        file_category: str,
        raw_file,
        file_meta,
        ids: str = None,
        source_type: Optional[str] = None,
        labels: Collection[AnyLabel] = (),
        gzip_compress_level: int = 5,
    ) -> dict:
        """Write file to S3"""
        if file_category not in WRITE_ALLOWED_CATEGORIES:
            raise InvalidFileCategoryError(
                f"{file_category} is not allowed category for write_file"
            )
        pattern = "(.*?)/(.*?)/(?:.*?)/(.*)"
        raw_file_key = re.sub("/{2,}", "/", raw_file["fileKey"])
        match = re.match(pattern, raw_file_key, flags=re.DOTALL)
        if not match:
            raise Exception(f'Raw file key {raw_file_key} does not match "{pattern}"')
        org_slug, destination_or_source_id, raw_file_path = match.groups()
        return self.write_detached_file(
            context=context,
            content=content,
            org_slug=org_slug,
            destination_or_source_id=destination_or_source_id,
            file_name=os.path.join(raw_file_path, file_name),
            file_category=file_category,
            file_meta=file_meta,
            ids=ids,
            source_type=source_type,
            labels=labels,
            gzip_compress_level=gzip_compress_level,
            bucket=raw_file["bucket"],
        )

    def write_detached_file(
        self,
        context,
        content: Union[str, bytes, io.BytesIO, Dict],
        org_slug: str,
        destination_or_source_id: str,
        file_name: str,
        file_category: str,
        file_meta,
        ids: str = None,
        source_type: Optional[str] = None,
        labels: Collection[AnyLabel] = (),
        gzip_compress_level: int = 5,
        bucket: str = os.environ.get("DATALAKE_BUCKET", None),
    ) -> dict:
        """Write file to S3"""
        disable_gzip = Datalake._should_disable_gzip(gzip_compress_level)

        if file_category == "IDS" and ids is None:
            raise Exception("ids can not be None when file_category is IDS")
        if source_type is not None:
            source_type_match = re.match("^[-a-z0-9]+$", source_type)
            if not source_type_match:
                raise Exception(
                    f'Source type "{source_type}" contains invalid character or upper case letter'
                )
        else:
            source_type = getOrEmptyString(file_meta, FIELDS["SOURCE_TYPE"], "unknown")

        file_key = os.path.join(
            org_slug, destination_or_source_id, file_category, file_name
        )

        is_valid_file_key, reason = self._is_file_path_valid(file_key)
        if not is_valid_file_key:
            raise InvalidPathException(file_key, reason)

        pipeline_config = context.get("pipelineConfig", {})
        pipeline_history_str = getUpdatedPipelineHistoryStr(
            getOrEmptyString(context, "pipelineId"),
            getOrEmptyString(file_meta, FIELDS["PIPELINE_HISTORY"]),
        )
        file_id = str(uuid4())
        ids_obj = VersionedRef(composite=ids)
        meta = {
            # constant
            FIELDS["INTEGRATION_TYPE"]: "datapipeline",
            FIELDS["VERSION"]: "2",
            # generated
            FIELDS["FILE_ID"]: file_id,
            # from raw file
            FIELDS["RAW_FILE_ID"]: getOrEmptyString(file_meta, FIELDS["FILE_ID"]),
            FIELDS["CUSTOM_METADATA"]: getOrEmptyString(
                file_meta, FIELDS["CUSTOM_METADATA"]
            ),
            FIELDS["CUSTOM_TAGS"]: getOrEmptyString(file_meta, FIELDS["CUSTOM_TAGS"]),
            FIELDS["SOURCE_ID"]: getOrEmptyString(file_meta, FIELDS["SOURCE_ID"]),
            FIELDS["DESTINATION_ID"]: getOrEmptyString(
                file_meta, FIELDS["DESTINATION_ID"]
            ),
            FIELDS["SOURCE_NAME"]: getOrEmptyString(file_meta, FIELDS["SOURCE_NAME"]),
            FIELDS["SOURCE_TYPE"]: source_type,
            FIELDS["TRACE_ID"]: getOrEmptyString(
                file_meta, FIELDS["TRACE_ID"], file_id
            ),
            # IDS/TMP
            **(
                {
                    FIELDS["IDS"]: ids_obj.composite,
                    FIELDS["IDS_TYPE"]: ids_obj.name,
                    FIELDS["IDS_VERSION"]: ids_obj.version,
                }
                if ids is not None and file_category in ["IDS", "TMP"]
                else {}
            ),
            # from pipeline context
            FIELDS["INTEGRATION_ID"]: getOrEmptyString(
                context, "pipelineId"
            ),  # pipeline id
            # https://github.com/tetrascience/ts-service-pipeline/blob/development/src/models/create-workflow-command.ts#L150
            FIELDS["INTEGRATION_NAME"]: getOrEmptyString(
                pipeline_config, "pipelineName"
            )[:254],
            FIELDS["PIPELINE_ID"]: getOrEmptyString(context, "pipelineId"),
            FIELDS["PIPELINE_WORKFLOW_ID"]: getOrEmptyString(context, "workflowId"),
            FIELDS["PIPELINE_MASTER_SCRIPT"]: (
                f"{context.get('masterScriptNamespace', '')}/{context.get('masterScriptSlug', '')}"
                f":{context.get('masterScriptVersion', '')}"
            ),
            FIELDS["PIPELINE_TASK_EXECUTION_ID"]: getOrEmptyString(context, "taskId"),
            FIELDS["PIPELINE_TASK_SCRIPT"]: getOrEmptyString(context, "taskScript"),
            FIELDS["PIPELINE_TASK_SLUG"]: getOrEmptyString(context, "taskSlug"),
            FIELDS["PIPELINE_HISTORY"]: pipeline_history_str,
            **(
                {FIELDS["API_USER_ID"]: os.getenv("TASK_USER_ID")}
                if os.getenv("TASK_USER_ID")
                else {}
            ),
        }

        if bucket is None:
            raise EnvironmentError(
                f"DATALAKE_BUCKET env or input file bucket is not set!"
            )

        params = {
            "Bucket": bucket,
            "Key": file_key,
            "Metadata": meta,
            "ServerSideEncryption": "aws:kms",
            "SSEKMSKeyId": f"alias/customer-key-{ENV}-{org_slug}",
        }

        if not disable_gzip:
            params["ContentEncoding"] = "gzip"

        if len(labels) > 0:
            self.create_labels_file(
                target_file={
                    "type": "s3file",
                    "bucket": bucket,
                    "fileKey": file_key,
                    "fileId": file_id,
                },
                labels=labels,
                sse_kms_key_id=f"alias/customer-key-{ENV}-{org_slug}",
            )

        if hasattr(content, "read"):
            # BytesIO object has .read() attribute
            response = S3FileobjUploader(
                self.s3,
                content,
                params,
                {"disable_gzip": disable_gzip, "compresslevel": gzip_compress_level},
            ).upload()
            # fakeS3 does not return VersionId, so use '' to avoid an exception
            version_id = response.get("VersionId", "")
            print({"level": "debug", "message": "file created", "response": response})
        else:
            # Otherwise `content` is either str, bytes or dict
            if isinstance(content, dict):
                if file_category == "IDS":
                    if self._ids_util is not None and "validate_ids" in self._ids_util:
                        self._ids_util["validate_ids"](
                            content, ids_obj.namespace, ids_obj.name, ids_obj.version
                        )
                    else:
                        raise MissingValidatorException
                    content = json.dumps(content, ignore_nan=True)
                else:
                    raise Exception("Writing non-IDS dicts to S3 is not supported")

            if not disable_gzip:
                if isinstance(content, str):
                    content = content.encode()
                content = gzip.compress(content, compresslevel=gzip_compress_level)

            if len(content) >= LARGE_FILE_SIZE_THRESHOLD_BYTES:
                print({"level": "debug", "message": "writing a 5+ GB file to s3"})

                # the upload_fileobj method requires a file like object that implements read. This
                # means we have to wrap content into an in memory stream. We will use BytesIO
                # instead of StringIO since the metadata has to be encoded. This means that if
                # content is a string, we have to convert it to bytes first.
                file_like_obj = None
                if isinstance(content, str):
                    file_like_obj = io.BytesIO(content.encode("utf-8"))
                else:
                    # content was already bytes so we should not try to encode it again
                    file_like_obj = io.BytesIO(content)
                try:
                    extra_args = self.extract_extra_args_from_params(params)

                    self.s3.upload_fileobj(
                        Fileobj=file_like_obj,
                        Bucket=bucket,
                        Key=file_key,
                        ExtraArgs=extra_args,
                    )
                    version_id = self.get_latest_obj_version(bucket, file_key)
                    print({"level": "debug", "message": "file created"})

                except Exception as exc:
                    raise Exception(
                        f"encountered an error updating the metadata because of {exc}"
                    ) from exc
                finally:
                    file_like_obj.close()
            else:
                response = self.s3.put_object(Body=content, **params)
                # fakeS3 does not return VersionId, so use '' to avoid an exception
                version_id = response.get("VersionId", "")
                print(
                    {"level": "debug", "message": "file created", "response": response}
                )

        result_file = {
            "type": "s3file",
            "bucket": bucket,
            "fileKey": file_key,
            "fileId": file_id,
            "version": version_id,
        }

        return result_file

    def update_metadata_tags(
        self,
        context: dict,
        file: dict,
        custom_meta: dict,
        custom_tags: list,
        options: dict = {},
    ):
        """Update S3 metadata tags"""
        bucket = file["bucket"]
        file_key = file["fileKey"]

        head = self.get_s3_head(file)
        current_meta = lowerMetadataKeys(head.get("Metadata"))

        if FIELDS["FILE_ID"] not in current_meta:
            raise KeyError(f"{FIELDS['FILE_ID']} not found in meta!")

        if (not custom_meta) and (not custom_tags):
            print(
                {
                    "level": "debug",
                    "message": "no action taken because no metadata or tags provided",
                }
            )
            return file

        is_ascii = (
            lambda s: s and isinstance(s, str) and bool(re.match(r"^[\x00-\x7F]*$", s))
        )

        custom_meta_str = current_meta.get(FIELDS["CUSTOM_METADATA"], "") or ""
        current_custom_meta = parse_query(custom_meta_str)
        if custom_meta:
            custom_meta_merged = {**current_custom_meta, **custom_meta}
            custom_meta_merged = {
                k: v for k, v in custom_meta_merged.items() if v is not None
            }
            for k, v in custom_meta_merged.items():
                if not is_ascii(k):
                    raise Exception(f"Metadata key {k} contains non-ASCII character")
                if not is_ascii(str(v)):
                    raise Exception(f"Metadata value {v} contains non-ASCII character")
            custom_meta_str = urlencode(custom_meta_merged)

        custom_tags_str = current_meta.get(FIELDS["CUSTOM_TAGS"], "")
        if custom_tags:
            custom_tags = list(map(str, custom_tags))
            for tag in custom_tags:
                if not is_ascii(tag):
                    raise Exception(f"Tag {tag} contains non-ASCII character")
            new_custom_tags = list(set(custom_tags_str.split(",") + custom_tags))
            new_custom_tags.sort()
            custom_tags_str = ",".join([t for t in new_custom_tags if t])

        if len(custom_meta_str) + len(custom_tags_str) >= 1024 * 1.5:
            raise Exception("Metadata and tags length larger than 1.5KB")

        new_file_id = options.get("new_file_id", str(uuid4()))

        pipeline_config = context.get("pipelineConfig", {})

        pipeline_history_str = getUpdatedPipelineHistoryStr(
            getOrEmptyString(context, "pipelineId"),
            getOrEmptyString(current_meta, FIELDS["PIPELINE_HISTORY"]),
        )

        params = {
            "Bucket": bucket,
            "CopySource": f"/{bucket}/{file_key}",
            # 'CopySourceIfUnmodifiedSince': head['LastModified'], # ensure no conflict?
            "Key": file_key,
            "ContentEncoding": head.get("ContentEncoding", None),
            "ContentType": head["ContentType"],
            "Metadata": {
                **current_meta,
                # constant
                FIELDS["INTEGRATION_TYPE"]: "datapipeline",
                FIELDS["VERSION"]: "2",
                FIELDS["FILE_ID"]: new_file_id,
                FIELDS["CUSTOM_METADATA"]: custom_meta_str,
                FIELDS["CUSTOM_TAGS"]: custom_tags_str,
                # Indicate that the content of the new file comes from previous file
                FIELDS["CONTENT_CREATED_FROM_FILE_ID"]: current_meta.get(
                    FIELDS["FILE_ID"]
                ),
                # Explicitly tell the TDP that this file will not inherit labels
                # from the source or pipeline input file (RAW FILE ID)
                FIELDS["DO_NOT_INHERIT_LABELS"]: "true",
                # from pipeline context
                FIELDS["INTEGRATION_ID"]: getOrEmptyString(context, "pipelineId"),
                FIELDS["INTEGRATION_NAME"]: getOrEmptyString(
                    pipeline_config, "pipelineName"
                )[:254],
                FIELDS["PIPELINE_ID"]: getOrEmptyString(context, "pipelineId"),
                FIELDS["PIPELINE_WORKFLOW_ID"]: getOrEmptyString(context, "workflowId"),
                FIELDS["PIPELINE_MASTER_SCRIPT"]: (
                    f"{context.get('masterScriptNamespace', '')}/"
                    f"{context.get('masterScriptSlug', '')}:"
                    f"{context.get('masterScriptVersion', '')}"
                ),
                FIELDS["PIPELINE_TASK_EXECUTION_ID"]: getOrEmptyString(
                    context, "taskId"
                ),
                FIELDS["PIPELINE_TASK_SCRIPT"]: getOrEmptyString(context, "taskScript"),
                FIELDS["PIPELINE_TASK_SLUG"]: getOrEmptyString(context, "taskSlug"),
                FIELDS["PIPELINE_HISTORY"]: pipeline_history_str,
            },
            "MetadataDirective": "REPLACE",
            "ServerSideEncryption": "aws:kms",
            "SSEKMSKeyId": head.get("SSEKMSKeyId", None),
        }

        params = {k: v for k, v in params.items() if v is not None}
        # fakeS3 does not return VersionId, so use '' to avoid an exception when we try to retrieve
        # from the head object or directly from the response object
        updated_version = ""

        # boto3.client.copy_object() has a maximum size of 5 GB. If the file exceeds this limit, we
        # have to use the boto3.copy() method that will do the multipart upload for us.
        file_size = head.get("ContentLength") or 0

        if file_size >= LARGE_FILE_SIZE_THRESHOLD_BYTES:
            print(
                {"level": "debug", "message": "updating the metadata of a 5+ GB file"}
            )
            extra_args = self.extract_extra_args_from_params(params)
            copy_source = {"Bucket": bucket, "Key": file_key}

            # unlike copy_object(), the copy() function does not return a response object. It will
            # raise an error instead. Additionally, we will have to make an additional request to
            # get the versionId if it exists
            try:
                self.s3.copy(
                    CopySource=copy_source,
                    Bucket=bucket,
                    Key=file_key,
                    ExtraArgs=extra_args,
                )
                updated_version = self.get_latest_obj_version(bucket, file_key)
            except Exception as exc:
                raise Exception(
                    f"encountered an error updating the metadata because of {exc}"
                ) from exc
        else:
            # For most cases when files are less than 5 GB in size. The versionId comes from the
            # response object of the copy_object() function
            response = self.s3.copy_object(**params)
            updated_version = response.get("VersionId", "")

        return {
            "type": "s3file",
            "bucket": bucket,
            "fileKey": file_key,
            "fileId": new_file_id,
            "version": updated_version,
        }

    def write_ids(
        self,
        context,
        content_obj: dict,
        file_suffix: str,
        raw_file,
        file_meta,
        ids: str,
        source_type: Optional[str],
        file_category: str,
        labels: Collection[AnyLabel],
        gzip_compress_level: int = 5,
    ) -> dict:
        """
        Write IDS to TDP using write_file.
        `content_obj` only accepts IDS content as a `dict`. This is required
        because `write_file` supports IDS validation only if the content
        passed to it is also a `dict`

        """
        ids_obj = VersionedRef(composite=ids)

        file_name = f"{ids_obj.namespace}-{ids_obj.name}-{file_suffix}"

        if not isinstance(content_obj, dict):
            raise TypeError(
                "'content_obj' passed to 'write_ids' must be of type 'dict'."
            )

        result = self.write_file(
            context,
            content_obj,
            file_name,
            file_category,
            raw_file,
            file_meta,
            ids,
            source_type,
            labels,
            gzip_compress_level,
        )
        return result

    def get_file_name(self, file: dict) -> str:
        """Get file basename from file metadata"""
        file_key = file["fileKey"]
        return os.path.basename(file_key)

    def get_presigned_url(self, file, ttl_sec):
        """Get Pre-signed URL from S3"""
        if ttl_sec is None or ttl_sec < 0 or ttl_sec > 900:
            raise Exception(
                "Cannot generate pre-signed S3 URL, expiration in seconds must be between 0 and "
                f"900, and specified value is {ttl_sec}"
            )

        kwargs = {"VersionId": file["version"]} if "version" in file else {}

        try:
            return self.s3.generate_presigned_url(
                "get_object",
                Params={
                    "Bucket": file["bucket"],
                    "Key": file["fileKey"],
                    **kwargs,
                },
                ExpiresIn=ttl_sec,
            )
        except Exception as exc:
            print(exc)

        return None

    def create_labels_file(
        self,
        target_file: dict,
        labels: Collection[AnyLabel],
        sse_kms_key_id: str = None,
    ):
        if sse_kms_key_id is None:
            try:
                head = self.get_s3_head(target_file)
                sse_kms_key_id = head.get("SSEKMSKeyId", "")
            except Exception as exc:
                print(exc)
                sse_kms_key_id = ""
        file_key = os.path.join(
            target_file["fileKey"], f'{target_file["fileId"]}.labels'
        )
        params = {
            "Bucket": target_file["bucket"],
            "Key": re.sub("/{2,}", "/", file_key)
            .replace("/RAW/", "/TMP/", 1)
            .replace("/IDS/", "/TMP/", 1)
            .replace("/PROCESSED/", "/TMP/", 1),
            "ContentType": "application/json",
        }
        if sse_kms_key_id != "":
            params["ServerSideEncryption"] = "aws:kms"
            params["SSEKMSKeyId"] = sse_kms_key_id
        else:
            params["ServerSideEncryption"] = "AES256"

        response = self.s3.put_object(
            Body=json.dumps(labels, cls=DataclassEncoder),
            **params,
        )
        print(
            {"level": "debug", "message": "labels file created", "response": response}
        )

    def extract_extra_args_from_params(self, params: dict) -> dict:
        """
        Return dictionary containing all arguments except Bucket, Key, CopySource

        These keys are already included in the copy() function signature.
        """
        return {
            k: v for k, v in params.items() if k not in ["Bucket", "Key", "CopySource"]
        }

    def get_latest_obj_version(self, bucket, file_key) -> str:
        """
        Get latest version of file_key in bucket

        We can't use s3.get_s3_head(file) because the file may have the old version before the copy
        command was called. So we need to make a direct call to head_object without the versionId
        to get the latest version of the file
        """
        try:
            new_head_obj = self.s3.head_object(Bucket=bucket, Key=file_key)
            return new_head_obj.get("VersionId", "")
        except Exception as exc:
            print(
                {
                    "level": "error",
                    "message": f"could not retrieve the latest version of the file because {exc}",
                }
            )
            return ""



from ts_sdk.task.log_codes.log_code import LogCode
from ts_sdk.task.log_codes.log_code_collection_meta import LogCodeCollectionMeta


class LogCodeCollection(metaclass=LogCodeCollectionMeta):
    # Platform interaction -- used only by the machinery that runs Task Scripts
    platform_interaction = LogCode(code_message="PlatformInteraction", code=100)
    # Informational/Generic/Uncategorized
    generic = LogCode(code_message="Generic", code=1000)
    # Input Data validation
    input_data_valid = LogCode(code_message="InputDataValid", code=1100)
    invalid_raw_input_data = LogCode(code_message="InvalidRawInputData", code=1101)
    invalid_input_file_attribute = LogCode(
        code_message="InvalidInputFileAttribute", code=1102
    )
    # Output Data Validation
    output_data_valid = LogCode(code_message="OutputDataValid", code=1200)
    invalid_ids_data = LogCode(code_message="InvalidIdsData", code=1201)
    invalid_output_data = LogCode(code_message="InvalidOutputData", code=1202)
    invalid_output_file_attribute = LogCode(
        code_message="InvalidOutputFileAttribute", code=1203
    )
    # Configuration
    #: Copy of the configuration (e.g. show inputs to task script)
    configuration = LogCode(code_message="Configuration", code=1300)
    invalid_pipeline_configuration = LogCode(
        code_message="InvalidPipelineConfiguration", code=1301
    )
    # Processing
    processing_status = LogCode(code_message="ProcessingStatus", code=1400)
    processing_begin = LogCode(code_message="ProcessingBegin", code=1401)
    processing_end = LogCode(code_message="ProcessingEnd", code=1402)
    processing_error = LogCode(code_message="ProcessingError", code=1403)
    space_odyssey = LogCode(code_message="SpaceOdyssey", code=2001)
    logger_error = LogCode(code_message="LoggerError", code=4900)



from threading import Timer

from .run import run_task_process
from .__util_log import Log
from .__util_task import (
    ContainerStoppedException,
    extend_task_timeout,
    poll_task,
    FailedPollTaskException,
)

log = Log({})


def healtcheck_worker(run_state):
    run_state["healtcheck_timer"] = None
    task = run_state["task"]
    task_process = run_state["task_process"]

    if task and task_process:
        task_id = task.get("id")
        try:
            extend_task_timeout(task)
        except:
            log.log(f"Error during timeout extension -> killing task {task_id}")
            task_process.kill()

    healtcheck_timer = Timer(60.0, healtcheck_worker, [run_state])
    run_state["healtcheck_timer"] = healtcheck_timer
    healtcheck_timer.start()


def main():
    run_state = {"task_process": None, "task": None, "healtcheck_timer": None}
    healtcheck_worker(run_state)

    def set_task_process_run_state(task_process):
        run_state["task_process"] = task_process

    while True:
        try:
            task = poll_task()
        except ContainerStoppedException:
            log.log("Container is stopped - exiting...")
            break
        except FailedPollTaskException:
            log.log("Cannot poll task - exiting...")
            break

        if task:
            task_id = task.get("id")
            log.log({"level": "debug", "message": f"Got new task {task_id}"})
            run_state["task"] = task
            shared_dict = run_task_process(task, set_task_process_run_state)
            log.log({"level": "debug", "message": f"Task {task_id} process completed"})
            run_state["task_process"] = None
            run_state["task"] = None

    if run_state["healtcheck_timer"] and run_state["healtcheck_timer"].is_alive():
        run_state["healtcheck_timer"].cancel()

    return shared_dict


if __name__ == "__main__":
    main()



import json
import re
import sys
import time
import traceback
from datetime import datetime

import requests
from colorama import Fore


def validate_dict_key(d, k):
    if d.get(k) is None:
        print(f"{Fore.RED}{k} is not set!{Fore.RESET}", file=sys.stderr, flush=True)
        return False
    return True


class TsApi:
    def __init__(self, **kwargs):
        res = validate_dict_key(kwargs, "org")
        res &= validate_dict_key(kwargs, "api_url")
        res &= validate_dict_key(kwargs, "auth_token")
        if not res:
            sys.exit(1)
        self.opts = kwargs

    @property
    def __api_url(self):
        return self.opts.get("api_url")

    @property
    def __request_defaults(self):
        return {
            "verify": self.opts.get("ignore_ssl") != True,
            "headers": self.__get_headers(),
        }

    def __get_headers(self):
        headers = {"x-org-slug": self.opts.get("org")}
        ts_auth = self.opts.get("auth_token")
        if re.compile(r"^([a-z0-9]+-)+[a-z0-9]+$").match(ts_auth, re.IGNORECASE):
            headers["x-api-key"] = ts_auth
        else:
            headers["ts-auth-token"] = ts_auth
        return headers

    def upload_artifact(self, cfg, artifact_bytes):
        type_to_url = {
            "ids": "ids",
            "master-script": "master-scripts",
            "protocol": "master-scripts",
            "task-script": "task-scripts",
        }
        url = f"{self.__api_url}/artifact/{type_to_url[cfg.type]}/{cfg.namespace}/{cfg.slug}/{cfg.version}"
        params = {
            **({"force": "true"} if cfg.force else {}),
        }
        r = requests.post(
            url,
            **self.__request_defaults,
            params=params,
            data=artifact_bytes,
        )
        if r.status_code < 400:
            return r.json()
        else:
            print(r.text, file=sys.stderr, flush=True)
            raise Exception(f"HTTP status: {r.status_code}, url: {r.url}")

    def get_task_script_build_info(self, id: str):
        url = f"{self.__api_url}/artifact/builds/{id}"
        r = requests.get(url, **self.__request_defaults)
        if r.status_code < 400:
            return r.json()
        else:
            print(r.text, file=sys.stderr, flush=True)
            raise Exception(f"HTTP status: {r.status_code}, url: {r.url}")

    def get_task_script_build_logs(self, id: str, params):
        url = f"{self.__api_url}/artifact/build-logs/{id}"
        r = requests.get(
            url,
            **self.__request_defaults,
            params={k: v for k, v in params.items() if v is not None},
        )
        if r.status_code < 400:
            return r.json()
        else:
            print(r.text, file=sys.stderr, flush=True)
            raise Exception(f"HTTP status: {r.status_code}, url: {r.url}")



from argparse import ArgumentParser

import colorama
from ts_sdk import __version__
from ts_sdk.task.__util_adapters.communication_format import (
    MAXIMUM_COMMUNICATION_FORMAT,
)

from .__init_cmd import init_cmd_args
from .__put_cmd import put_cmd_args
from .__utils import check_update_required


def main():
    colorama.init()

    parser = ArgumentParser(
        prog="ts-sdk",
        description="This cli is being deprecated May 1st, 2025. Try out the new cli: `pip3 install tetrascience-cli`",
    )

    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--max-supported-communication-format",
        action="version",
        version=MAXIMUM_COMMUNICATION_FORMAT.value,
        help="TDP communication format",
    )

    subparsers = parser.add_subparsers()

    init_cmd_args(
        subparsers.add_parser(
            "init", help="initialize master and task script from a template"
        )
    )

    put_cmd_args(
        subparsers.add_parser(
            "put", help="puts artifact identified by namespace/slug:version"
        )
    )

    args = parser.parse_args()

    check_update_required(__version__)

    if hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import io
import os
import pathlib
import tempfile
import typing as t

import typing_extensions as te

from ..task.__util_config import (
    MISSING_ALLOWED_IDS,
    AllowedIDS,
    FunctionConfig,
    MissingAllowedIDS,
)
from ..task.__util_log import Log
from ..task.data_model import File, FileCategory, ReadResult


class Context:
    """A development-time version of the context object that is passed into
    the task script handler when running as part of a pipeline.
    """

    def __init__(self, slug: t.Optional[str] = None, pipeline_config={}):
        self._storage = {}
        self._pipeline_config = pipeline_config
        self._log = Log({})
        self.slug = slug
        self._allowed_ids = self._get_allowed_ids()

    @property
    def allowed_ids(self):
        if isinstance(self._allowed_ids, MissingAllowedIDS):
            raise KeyError("'allowedIds' is not present in config.json")

        return self._allowed_ids

    def _get_allowed_ids(self) -> t.Union[MissingAllowedIDS, AllowedIDS]:
        if self.slug:
            func_dir = pathlib.Path(os.getcwd())
            func_config = FunctionConfig.from_task_script_dir(func_dir, self.slug)
            return func_config.allowed_ids
        return MISSING_ALLOWED_IDS

    @property
    def pipeline_config(self) -> t.Dict[str, str]:
        """Pipeline configuration including secrets."""
        return self._pipeline_config

    def read_file(self, file: File, form: str = "body") -> ReadResult:
        if form == "body":
            return self._storage[file["fileKey"]]
        elif form == "file_obj":
            result = self._storage[file["fileKey"]]
            body = result["body"]
            buf = io.BytesIO()
            buf.write(body)
            buf.seek(0)
            result["file_obj"] = buf
            return result
        elif form == "download":
            result = self._storage[file["fileKey"]]
            with tempfile.NamedTemporaryFile(mode="wb", delete=False) as tf:
                tf.write(result["body"])
                result["download"] = tf.name
            return result
        raise Exception(f"Invalid form: {form}")

    def write_file(
        self,
        content: t.AnyStr,
        file_name: str,
        file_category: FileCategory,
        ids: str = None,
        custom_metadata: t.Dict[str, str] = None,
        custom_tags: t.List[str] = None,
        source_type: str = None,
        labels: t.Iterable[t.Mapping[te.Literal["name", "value"], str]] = [],
    ) -> File:
        if (
            self.slug
            and file_category == "IDS"
            and self._allowed_ids != MISSING_ALLOWED_IDS
        ):
            ids = self.allowed_ids.get_ids_to_be_written(ids)

        if type(content) == str:
            content = content.encode("UTF-8")

        self._storage[file_name] = {
            "metadata": {
                "TS_IDS": ids,
                "TS_SOURCE_TYPE": source_type,
                "TS_FILE_CATEGORY": file_category,
            },
            "body": content,
            "custom_metadata": custom_metadata,
            "custom_tags": custom_tags,
        }
        return {
            "type": "s3file",
            "bucket": "fake-unittest-bucket",
            "fileKey": file_name,
        }

    def get_ids(self, namespace: str, slug: str, version: str):
        return {}

    # always return true in local context
    def validate_ids(self, data: dict, namespace: str, slug: str, version: str) -> bool:
        if self.slug and self._allowed_ids != MISSING_ALLOWED_IDS:
            ids_str = f"{namespace}/{slug}:{version}"
            self.allowed_ids.is_reconcilable(ids_str)

        return True

    def write_ids(
        self,
        content_obj,
        file_suffix: str,
        ids: t.Optional[str] = None,
        custom_metadata: t.Mapping[str, str] = {},
        custom_tags: t.Iterable[str] = [],
        source_type: t.Optional[str] = None,
        file_category: t.Optional[str] = "IDS",
        labels: t.Iterable[t.Mapping[te.Literal["name", "value"], str]] = [],
    ) -> File:
        if file_category not in ["TMP", "IDS"]:
            file_category = "IDS"

        if (
            self.slug
            and file_category == "IDS"
            and self._allowed_ids != MISSING_ALLOWED_IDS
        ):
            ids = self.allowed_ids.get_ids_to_be_written(ids)

        return {}

    def get_file_name(self, file: File) -> str:
        return ""

    def get_logger(self):
        return self._log

    def get_secret_config_value(self, secret_name: str, silent_on_error=True) -> str:
        return self._pipeline_config[secret_name]

    def resolve_secret(self, secret) -> str:
        return secret

    def get_presigned_url(self, file: File, ttl_sec=300) -> str:
        return ""

    def update_metadata_tags(
        self,
        file: File,
        custom_meta: t.Mapping[str, str] = {},
        custom_tags: t.Iterable[str] = [],
    ) -> File:
        return {}

    def run_command(
        self,
        org_slug,
        target_id,
        action,
        metadata,
        payload,
        ttl_sec=300,
        initial_delay_sec=0,
    ):
        return {}

    def add_labels(self, file, labels):
        return {}

    def get_labels(self, file):
        return {}

    def delete_labels(self, file):
        return {}



from .make_adapter import (
    make_adapter,
    select_versioned_value,
    __make_implementation_from_formats,
)
from .curry_format import curry_for_each_method
from .communication_format import (
    DEFAULT_COMMUNICATION_FORMAT,
    UNKNOWN_COMMUNICATION_FORMAT,
    CommunicationFormat,
)



import functools
import importlib
import json
import os
import re
import sys
import typing as t
import uuid

from ts_sdk.task.__exceptions import FatalError
from ts_sdk.task.__util_fileinfo import (
    add_labels,
    delete_labels,
    get_file_pointer,
    get_labels,
)

from .__util_adapters import CommunicationFormat, select_versioned_value
from .__util_adapters.communication_format import get_communication_format
from .__util_command import run_command
from .__util_config import (
    MISSING_ALLOWED_IDS,
    AllowedIDS,
    FunctionConfig,
    IDSValidateWrongParams,
    MissingAllowedIDS,
    NoAllowedIdsSpecified,
)
from .__util_datalake import Datalake
from .__util_decorators import deprecated
from .__util_es_datalake import es_datalake_search_eql, es_hit_to_file_pointer
from .__util_ids import create_ids_util
from .__util_log import Log
from .__util_merge import merge_arrays, merge_objects
from .__util_metadata import FIELDS
from .__util_permalink import ObjectType, get_permalink
from .__util_storage import Storage
from .__util_validation import (
    validate_file_labels,
    validate_file_meta,
    validate_file_tags,
)
from .data_model import AnyLabel, File, FileCategory, ReadResult

COMPLETED = "completed"
FAILED = "failed"

DEFAULT_GZIP_COMPRESS_LEVEL = int(os.environ.get("DEFAULT_GZIP_COMPRESS_LEVEL", 5))

# TODO these should be logged at a more verbose level than debug
# LOG_TAG_PRE_FUNCTION = 'pre_function_call'
# LOG_TAG_POST_FUNCTION = 'post_function_call'
LOG_TAG_SCRIPT_STARTED = "script_started"
LOG_TAG_SCRIPT_ENDED = "script_ended"

if "default_print" not in __builtins__:
    __builtins__["default_print"] = __builtins__["print"]


def wrap_log(func_name):
    def return_wrapper(fn):
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            # TODO log at a finer level
            # id = str(uuid.uuid4())
            # Context.log.log({
            #     'level': 'debug',
            #     'tag': LOG_TAG_PRE_FUNCTION,
            #     'funcName': func_name,
            #     'id': id
            # })
            result = fn(*args, **kwargs)
            # Context.log.log({
            #     'level': 'debug',
            #     'tag': LOG_TAG_POST_FUNCTION,
            #     'funcName': func_name,
            #     'id': id
            # })
            return result

        return wrapper

    return return_wrapper


def camel_to_snake(value: str) -> str:
    return "".join(["_" + i.lower() if i.isupper() else i for i in value]).lstrip("_")


class Context:
    """A context object that is passed into
    the task script handler when running as part of a pipeline.
    """

    org_slug: str
    pipeline_id: str
    workflow_id: str

    master_script_namespace: str
    master_script_slug: str
    master_script_version: str

    input_file: File

    created_at: str
    task_id: str
    task_created_at: str

    pipeline_config: t.Mapping[str, str]

    platform_url: str
    platform_api_url: str
    platform_version: str

    tmp_dir: str

    def __init__(
        self,
        obj,
        datalake,
        ids_util,
        log,
        allowed_ids: t.Union[AllowedIDS, MissingAllowedIDS] = MISSING_ALLOWED_IDS,
    ):
        self._obj = {
            **obj,
        }  # keys are later converted to snake case via "camel_to_snake"
        for key in self._obj:
            setattr(self, camel_to_snake(key), self._obj[key])
        self._datalake = datalake
        self._ids_util = ids_util
        self._log = log
        self._allowed_ids = allowed_ids
        self.tmp_dir = os.environ.get("TMPDIR")

    @property
    def allowed_ids(self) -> AllowedIDS:
        """Returns AllowedIDS object that either stores the list of allowed ids or
        a dict of allowed ids in `allowed_ids` attribute. These values are populated
        from `config.json`

        Raises:
            NoAllowedIdsSpecified: Raised if `config.json` has no `allowedIds` key
            for the task-script's slug.

        Returns:
            AllowedIDS object
        """
        if self._allowed_ids is MISSING_ALLOWED_IDS:
            raise NoAllowedIdsSpecified(
                "No 'allowedIds' found for the function being called. Please check 'config.json'."
            )
        return self._allowed_ids

    @wrap_log("context.get_file_pointer")
    def get_file_pointer(self, file_id: str) -> File:
        return get_file_pointer(self._obj, file_id)

    @wrap_log("context.read_file")
    def read_file(self, file: File, form: str = "body") -> ReadResult:
        """Reads a file from the data lake and returns its contents in one of
        three forms.

        If form='body' (the default), then result['body'] holds the contents of
        the file as a byte array. This approach cannot handle large files that
        don't fit in memory.

        If form='file_obj', then result['file_obj'] is a file-like object that
        can be used to access the body in a streaming manner. This object can
        be passed to Python libraries such as Pandas.

        If form='download', then result['download'] is the file name of a local
        file that has been downloaded from the specified data lake file. This
        is useful when the data needs to be processed by native code (e.g.
        SQLite) or an external utility program.

        >>> import json
        >>> import pandas as pd
        >>> import sqlite3
        >>> def task(input, context):
        ...     f = context.read_file(input, form='body')
        ...     json.loads(f['body'])
        ...
        ...     f = context.read_file(input, form='file_obj')
        ...     df = pd.read_csv(f['file_obj'])
        ...
        ...     f = context.read_file(input, form='download')
        ...     con = sqlite3.connect(f['download'])
        ...     df = pd.read_sql_query('SELECT * FROM foo', con)
        """

        if len(file.keys()) == 1 and file["fileId"] is not None:
            es_files = self.search_eql(
                {"size": 1, "query": {"term": {"fileId": {"value": file["fileId"]}}}},
                returns="filePointers",
            )
            if len(es_files) == 0:
                raise Exception(f"File with fileId {file['fileId']} not found!")
            file = es_files[0]

        return self._datalake.read_file(file, form, self.tmp_dir)

    @wrap_log("context.write_file")
    def write_file(
        self,
        content: t.Union[bytes, t.BinaryIO, str, t.Dict],
        file_name: str,
        file_category: FileCategory,
        ids: t.Optional[str] = None,
        custom_metadata: t.Optional[t.Mapping[str, str]] = None,
        custom_tags: t.Iterable[str] = (),
        source_type: t.Optional[str] = None,
        labels: t.Collection[AnyLabel] = (),
        gzip_compress_level: int = DEFAULT_GZIP_COMPRESS_LEVEL,
    ) -> File:
        """
        Writes output object to the data lake

        If content is a dictionary is passed with "IDS" file_category, IDS validation will be
        performed before serializing and writing dict to data lake.

        If `allowedIds` is defined for the function in `config.json` then IDS file will be
        created only if `ids` is allowed, i.e.:
        1. if `ids` passed is present as `allowedIds` for the function slug in config.json
        2. if `ids` passed is `None` and `allowedIDS` for function slug list only one IDS, then it
        will be used for writing the IDS file.
        An error will be raised if:
        1. `allowedIds` is not defined for the function slug and the value of the `ids` parameter is `None`
        2. `allowedIds` is set to `null` for the function slug (since that indicates the function does not write IDSes)

        Arguments:
            content: Data to be written, can be a file-like object, string or an IDS dictionary
            file_name: Filename of the object to be written
            file_category: Category of file to be written
            ids: (Optional) IDS value in format of NAMESPACE/SLUG:VERSION
            custom_metadata: (Optional) Additional metadata to be written
            custom_tags: (Optional) Additional tags to be written
            source_type: (Optional) Source type of object being written
            labels: (Optional) Labels for object being written
            gzip_compress_level: (Optional) 1 is fastest and 9 is slowest. 0 is no compression. The default is 5
        """

        if file_category == "IDS" and self._allowed_ids != MISSING_ALLOWED_IDS:
            ids = self.allowed_ids.get_ids_to_be_written(
                ids_requested_to_be_written=ids
            )

        if custom_metadata is None:
            custom_metadata = {}
        raw_file = self.input_file
        validate_file_meta(custom_metadata)
        validate_file_tags(custom_tags)
        validate_file_labels(labels)
        file_meta = {
            # in case custom_metadata & custom_tags are undefined in raw_file meta
            FIELDS["CUSTOM_METADATA"]: "",
            FIELDS["CUSTOM_TAGS"]: "",
            **self._datalake.get_file_meta(raw_file),
        }
        file_meta[FIELDS["CUSTOM_METADATA"]] = merge_objects(
            file_meta.get(FIELDS["CUSTOM_METADATA"], ""), custom_metadata
        )
        file_meta[FIELDS["CUSTOM_TAGS"]] = merge_arrays(
            file_meta.get(FIELDS["CUSTOM_TAGS"], ""),
            custom_tags,
        )
        file_meta[FIELDS["TRACE_ID"]] = file_meta.get(
            FIELDS["TRACE_ID"], raw_file["fileId"]
        )
        return self._datalake.write_file(
            context=self._obj,
            content=content,
            file_name=file_name,
            file_category=file_category,
            raw_file=raw_file,
            file_meta=file_meta,
            ids=ids,
            source_type=source_type,
            labels=labels,
            gzip_compress_level=gzip_compress_level,
        )

    @wrap_log("context.write_detached_file")
    def write_detached_file(
        self,
        content: t.Union[bytes, t.BinaryIO, str, t.Dict],
        file_name: str,
        file_category: FileCategory,
        custom_metadata: t.Optional[t.Mapping[str, str]] = None,
        custom_tags: t.Iterable[str] = (),
        source_type: t.Optional[str] = "pipeline",
        source_name: t.Optional[str] = None,
        source_id: t.Optional[str] = None,
        labels: t.Collection[AnyLabel] = (),
        gzip_compress_level: int = DEFAULT_GZIP_COMPRESS_LEVEL,
    ) -> File:
        """
        Writes an object to the data lake (detached)

        Arguments:
            content: Data to be written, can be a file-like object, string or an IDS dictionary
            file_name: Filename of the object to be written
            file_category: Category of file to be written
            custom_metadata: (Optional) Additional metadata to be written
            custom_tags: (Optional) Additional tags to be written
            source_type: (Optional) Source type of object being written
            source_name: (Optional) Source name of object being written
            source_id: (Optional) Source id of object being written
            labels: (Optional) Labels for object being written
            gzip_compress_level: (Optional) 1 is fastest and 9 is slowest. 0 is no compression. The default is 5
        """
        if custom_metadata is None:
            custom_metadata = {}
        validate_file_meta(custom_metadata)
        validate_file_tags(custom_tags)
        validate_file_labels(labels)
        file_meta = {
            FIELDS["CUSTOM_METADATA"]: merge_objects("", custom_metadata),
            FIELDS["CUSTOM_TAGS"]: merge_arrays("", custom_tags),
            FIELDS["SOURCE_TYPE"]: source_type,
            FIELDS["SOURCE_NAME"]: (
                source_name
                if source_name is not None
                else self.pipeline_config.get("pipelineName", "")
            ),
        }
        return self._datalake.write_detached_file(
            context=self._obj,
            content=content,
            org_slug=self._obj["orgSlug"],
            destination_or_source_id=(
                source_id if source_id is not None else self._obj["pipelineId"]
            ),
            file_name=file_name,
            file_category=file_category,
            file_meta=file_meta,
            labels=labels,
            gzip_compress_level=gzip_compress_level,
        )

    @wrap_log("context.get_ids")
    def get_ids(self, namespace: str, slug: str, version: str):
        """Returns IDS schema"""
        return self._ids_util["get_ids"](namespace, slug, version)

    @wrap_log("context.validate_ids")
    def validate_ids(
        self,
        data: t.Any,
        namespace: t.Optional[str] = None,
        slug: t.Optional[str] = None,
        version: t.Optional[str] = None,
        ignore_allowed_ids: bool = False,
    ):
        """
        1. Determine IDS from allowedIds when namespace, slug and versions are None
        2. Check if IDS is present in 'allowedIds', when `allowedIds` is defined in `config.json`.
        3. Checks validity of IDS content provided in `data`.
        Throws an error if not valid or not possible to determine IDS from allowedIds
        """

        if namespace and slug and version:
            pass
        elif namespace or slug or version:
            raise IDSValidateWrongParams(
                f"Namespace, slug and version have to be set or equal to None: {namespace}/{slug}:{version}"
            )
        else:
            # all namespace, slug and version are None
            if self._allowed_ids == MISSING_ALLOWED_IDS:
                raise IDSValidateWrongParams(
                    f"Not possible to determine the IDS to validate against: allowedIds is missing"
                )

            if self._allowed_ids.is_single():
                namespace, slug, version = self._allowed_ids.get_first().to_tuple()
            else:
                raise IDSValidateWrongParams(
                    f"Not possible to determine the IDS to validate against: allowedIds has more than 1 value"
                )

        if not ignore_allowed_ids and self._allowed_ids != MISSING_ALLOWED_IDS:
            ids_str = f"{namespace}/{slug}:{version}"
            self.allowed_ids.is_reconcilable(ids_str)

        return self._ids_util["validate_ids"](data, namespace, slug, version)

    @wrap_log("context.write_ids")
    def write_ids(
        self,
        content_obj: t.Union[bytes, t.BinaryIO, str, t.Dict],
        file_suffix: str,
        ids: t.Optional[str] = None,
        custom_metadata: t.Optional[t.Mapping[str, str]] = None,
        custom_tags: t.Iterable[str] = (),
        source_type: t.Optional[str] = None,
        file_category: t.Optional[FileCategory] = "IDS",
        labels: t.Collection[AnyLabel] = (),
        gzip_compress_level: int = DEFAULT_GZIP_COMPRESS_LEVEL,
    ) -> File:
        """
        Writes output IDS to the data lake.  If `content_obj` is a dictionary, then it will also be validated against
        the IDS schema.
        If `allowedIds` is defined for the function in `config.json`, then the IDS file will be created only if `ids`
        is allowed ie:
        1. if `ids` passed is present as `allowedIds` for the function slug in config.json
        2. if `ids` passed is `None` and `allowedIDS` for function slug list only one IDS, then it
        will be used for writing the IDS file.
        An error will be raised if:
        1. `allowedIds` is not defined for the function slug and the value of the `ids` parameter is `None`
        2. `allowedIds` is set to `null` for the function slug (since that indicates the function does not write IDSes)

        Arguments:
            content_obj: Data to be written, can be a file-like object, string or an IDS dictionary
            file_suffix: IDS filenames follow the form {IDS_NAMESPACE}-{IDS_SLUG}-{file_suffix}.json
            ids: The IDS to use for validation.  String of the form NAMESPACE/SLUG:VERSION.  If `None` (default), an
                Exception is raised.
            custom_metadata: (Optional) Additional metadata to be written
            custom_tags: (Optional) Additional tags to be written
            source_type: (Optional) Source type of object being written.  If `None` (default), value comes from the
                value in `custom_metadata` belonging to the key "ts_source_type".  If the key does not exist, it
                defaults to "unknown"
            file_category: Category of file to be written. Defaults to "IDS".  If not "IDS" or "TMP", the category will
                be "IDS".
            labels: (Optional) Labels for object being written
            gzip_compress_level: (Optional) 1 is fastest and 9 is slowest. 0 is no compression. The default is 5
        """
        if file_category not in ("IDS", "TMP"):
            file_category = "IDS"

        if file_category == "IDS" and self._allowed_ids != MISSING_ALLOWED_IDS:
            ids = self.allowed_ids.get_ids_to_be_written(
                ids_requested_to_be_written=ids
            )

        if custom_metadata is None:
            custom_metadata = {}
        raw_file = self.input_file
        validate_file_meta(custom_metadata)
        validate_file_tags(custom_tags)
        validate_file_labels(labels)
        file_meta = {
            # in case custom_metadata & custom_tags are undefined in raw_file meta
            FIELDS["CUSTOM_METADATA"]: "",
            FIELDS["CUSTOM_TAGS"]: "",
            **self._datalake.get_file_meta(raw_file),
        }
        file_meta[FIELDS["CUSTOM_METADATA"]] = merge_objects(
            file_meta.get(FIELDS["CUSTOM_METADATA"], ""), custom_metadata
        )
        file_meta[FIELDS["CUSTOM_TAGS"]] = merge_arrays(
            file_meta.get(FIELDS["CUSTOM_TAGS"], ""),
            custom_tags,
        )
        file_meta[FIELDS["TRACE_ID"]] = file_meta.get(
            FIELDS["TRACE_ID"], raw_file["fileId"]
        )

        return self._datalake.write_ids(
            context=self._obj,
            content_obj=content_obj,
            file_suffix=file_suffix,
            raw_file=raw_file,
            file_meta=file_meta,
            ids=ids,
            source_type=source_type,
            file_category=file_category,
            labels=labels,
            gzip_compress_level=gzip_compress_level,
        )

    def get_file_name(self, file: File) -> str:
        """Returns the filename of the file not downloading it locally"""
        return self._datalake.get_file_name(file)

    @wrap_log("context.get_file_permalink")
    def get_file_permalink(self, file: File) -> str:
        return get_permalink(
            self.platform_api_url, ObjectType.FILE, self.get_file_id(file)
        )

    def get_logger(self):
        """Returns the structured logger object.
        The input should be an object (eg. containing a message field, among others).

        ...     logger = context.get_logger()
        ...     logger.log({
        ...         "message": "Starting the main parser",
        ...         "level": "info"
        ...     })

        """
        return self._log

    @deprecated(
        "'context.get_secret_config_value' deprecated, use 'context.resolve_secret' instead!"
    )
    def get_secret_config_value(self, secret_name: str, silent_on_error=True) -> str:
        """Returns the value of the secret.
        If secret is missing, empty string or throws error, depending on the second argument.
        """

        return get_secret_config_value(self._obj, secret_name, silent_on_error)

    def resolve_secret(self, secret) -> t.Optional[str]:
        """Returns the value of the secret."""

        if type(secret) is dict and "ssm" in secret:
            pipeline_config = self._obj.get("pipelineConfig")
            if f'_resolved_:{secret["ssm"]}' in pipeline_config:
                return pipeline_config.get(f'_resolved_:{secret["ssm"]}')
            key = re.sub(r"^/[^/]*/[^/]*/org-secrets/", "", secret["ssm"])
            key = re.sub(r"[^a-z0-9]+", "_", key, flags=re.IGNORECASE)
            secret_value = os.environ.get("SECRET_" + key)
            return secret_value
        return secret

    def get_presigned_url(self, file: File, ttl_sec=300) -> str:
        """Returns a time-limited HTTPS URL that can be used to access the file.
        If URL generation fails for any reason (except invalid value for ttl_sec parameter) `None` will be returned.
        """
        return self._datalake.get_presigned_url(file, ttl_sec)

    @wrap_log("context.update_metadata_tags")
    def update_metadata_tags(
        self,
        file: File,
        custom_meta: t.Optional[t.Mapping[str, str]] = None,
        custom_tags: t.Iterable[str] = (),
        options: t.Mapping[str, str] = {},
    ) -> File:
        """Updates file's custom metadata and tags.
        Use 'None' to remove a meta entry.
        New tags will be appended to existing ones.
        """
        if custom_meta is None:
            custom_meta = {}
        validate_file_meta(custom_meta)
        validate_file_tags(custom_tags)
        return self._datalake.update_metadata_tags(
            context=self._obj,
            file=file,
            custom_meta=custom_meta,
            custom_tags=custom_tags,
            options=options,
        )

    @deprecated("'context.run_command' deprecated, use 'context.run_cmd' instead!")
    @wrap_log("context.run_command")
    def run_command(
        self,
        org_slug,
        target_id,
        action,
        metadata,
        payload,
        ttl_sec=300,
        initial_delay_sec=0,
        return_command=False,
    ):
        """Invokes remote command/action on target (agent or connector) and returns its response"""
        return run_command(
            self._obj,
            org_slug,
            target_id,
            action,
            metadata,
            payload,
            ttl_sec,
            initial_delay_sec,
            return_command,
        )

    @wrap_log("context.run_cmd")
    def run_cmd(
        self,
        target_id,
        action,
        metadata,
        payload,
        ttl_sec=300,
        initial_delay_sec=0,
        return_command=False,
    ):
        """Invokes remote command/action on target (agent or connector) and returns its response"""
        return run_command(
            self._obj,
            self.org_slug,
            target_id,
            action,
            metadata,
            payload,
            ttl_sec,
            initial_delay_sec,
            return_command,
        )

    def get_file_id(self, file):
        if "fileId" in file:
            file_id = file["fileId"]
        else:
            file_metadata = self._datalake.get_file_meta(file)
            file_id = file_metadata.get(FIELDS["FILE_ID"])
        return file_id

    @wrap_log("context.add_labels")
    def add_labels(
        self,
        file: File,
        labels: t.Optional[t.Collection[AnyLabel]],
        no_propagate: bool = False,
    ):
        if validate_file_labels(labels):
            file_id = self.get_file_id(file)
            return add_labels(self._obj, file_id, labels, no_propagate)
        else:
            print(
                {
                    "level": "warning",
                    "message": "no labels provided in add_labels()!",
                }
            )
            return self.get_labels(file)

    def get_labels(self, file):
        file_id = self.get_file_id(file)
        return get_labels(self._obj, file_id)

    @wrap_log("context.delete_labels")
    def delete_labels(self, file, label_ids):
        file_id = self.get_file_id(file)
        return delete_labels(self._obj, file_id, label_ids if label_ids else [])

    @wrap_log("context.add_attributes")
    def add_attributes(
        self,
        file: File,
        custom_meta: t.Optional[t.Mapping[str, str]] = None,
        custom_tags: t.Iterable[str] = (),
        labels: t.Collection[AnyLabel] = (),
    ) -> File:
        if custom_meta is None:
            custom_meta = {}

        # case where only labels are provided
        if not custom_meta and not custom_tags:
            if labels:
                # no need to validate labels before here, because add_labels will validate for us
                self.add_labels(file, labels)
            else:
                print(
                    {
                        "level": "warning",
                        "message": "no attributes provided in add_attributes()!",
                    }
                )
            return file

        # case where meta and/or tags are specified

        new_file_id = str(uuid.uuid4())

        if validate_file_labels(labels):
            self._datalake.create_labels_file(
                target_file={**file, "fileId": new_file_id},
                # TODO: this line will overwrite the label, instead, should merge with the existing labels
                # otherwise, the behavior is different from when metadata, tag are not provided.
                labels=labels,
            )

        # metadata and tags are validated in this function
        new_file = self.update_metadata_tags(
            file=file,
            custom_meta=custom_meta,
            custom_tags=custom_tags,
            options={"new_file_id": new_file_id},
        )

        return new_file

    @wrap_log("context.search_eql")
    def search_eql(self, payload, returns="raw", query=None):
        search_res = es_datalake_search_eql(payload, query=query)
        if returns == "filePointers":
            return list(map(es_hit_to_file_pointer, search_res["hits"]["hits"]))
        return search_res

    @wrap_log("context.raise_fatal_error")
    def raise_fatal_error(self, message: str = "No message"):
        # From CommunicationFormatV2 onward, the platform supports fatal errors
        if select_versioned_value(
            {
                CommunicationFormat.V0: True,
                CommunicationFormat.V2: False,
            }
        ):
            self.get_logger().log(
                "Fatal errors are not supported in this environment, and will be treated like a regular error"
            )
        raise FatalError(message)


def output_response(storage, response, correlation_id):
    storage.writeObject({**response, "id": correlation_id})


def resolve_func(func_dir, func_slug):
    func_conf_file = os.path.join(func_dir, "config.json")
    with open(func_conf_file, "r") as file:
        func_conf = json.load(file)
    for f in func_conf["functions"]:
        if f["slug"] == func_slug:
            function = f["function"]
            break
    else:
        raise Exception(f"function not found: {func_slug}")
    # print(function)
    func_module, _, func_name = function.rpartition(".")
    return func_module, func_name


def resolve_secrets_in_pipeline_config(context_from_arg):
    secrets = {}
    pipeline_config = context_from_arg.get("pipelineConfig")
    for key in pipeline_config:
        if key.startswith("ts_secret_name_"):
            secret_name = key.split("ts_secret_name_", 1)[1]
            secrets[secret_name] = get_secret_config_value(context_from_arg, key, True)
    return secrets


def get_secret_config_value(context_from_arg, secret_name, silent_on_error=True):
    pipeline_config = context_from_arg.get("pipelineConfig")

    if secret_name.startswith("ts_secret_name_"):
        secret_short_name = secret_name.split("ts_secret_name_", 1)[1]
        secret_full_key = secret_name
    else:
        secret_short_name = secret_name
        secret_full_key = "ts_secret_name_" + secret_name

    if f"_resolved_:{secret_short_name}" in pipeline_config:
        return pipeline_config.get(f"_resolved_:{secret_short_name}")

    if (
        not secret_full_key in pipeline_config.keys()
        or pipeline_config.get(secret_full_key) is None
    ):
        if silent_on_error:
            Context.log.log(
                f"Secret {secret_full_key} not found in the workflow config."
            )
            return ""
        else:
            raise Exception(
                f"Secret {secret_full_key} not found in the workflow config."
            )

    try:
        if os.environ.get("TASK_SCRIPTS_CONTAINERS_MODE") == "ecs":
            secret_value = os.environ.get(
                "TS_SECRET_"
                + re.sub(r"[^a-z0-9]+", "_", secret_short_name, flags=re.IGNORECASE)
            )
            if secret_value is None:
                raise Exception(f"Secret {secret_short_name} not found")
            return secret_value
        raise Exception(f"Could not resolve secret value for  {secret_full_key}")
    except Exception as e:
        if silent_on_error:
            Context.log.log(e)
            return ""
        else:
            raise Exception(
                f"Could not resolve secret value for  {secret_full_key}"
            ) from e


def should_override_builtin_print():
    """
    This is the essence of CommunicationFormat v2
    If v2: don't override the built-in print, because we know
    that the fluentbit sidecar is present, and will capture all
    stdout and send it to s3/cloudwatch
    :return:
    """
    return select_versioned_value(
        {
            CommunicationFormat.V0: True,
            CommunicationFormat.V2: False,
        }
    )


def construct_shadow_print(log: Log):
    """
    Constructs a function that captures `log`
    Intended to replace the __builtin__ print statement and pass the arguments to Log class
    which will decorate log messages with more information
    :param log:
    :return:
    """

    def shadow_print(*values, sep=" ", end="\n", file=sys.stdout, flush=False):
        """
        If a file is not sys.stdout:
            uses the default print because a user is attempting to print to a file
        Otherwise, defers to Log.log
        """
        if file is not sys.stdout:
            __builtins__["default_print"](
                *values, sep=sep, end=end, file=file, flush=flush
            )
        else:
            # No need to pass flush to our logger, the script's dockerfile is set up to always flush stdout
            log.log(*values, sep=sep)

    return shadow_print


def run(
    input,
    context_from_arg,
    func,
    correlation_id,
    func_dir,
    storage_type,
    storage_bucket,
    storage_file_key,
    storage_endpoint,
    artifact_bucket,
    artifact_prefix,
    artifact_endpoint,
    artifact_file_key,
    artifact_bucket_private,
    artifact_prefix_private,
    artifact_endpoint_private,
    store_output=True,
):
    log = Log(context_from_arg)
    Context.log = log
    log.log({"level": "debug", "tag": LOG_TAG_SCRIPT_STARTED})
    log.log(
        {
            "level": "debug",
            "communication format version": get_communication_format().value,
        }
    )
    if storage_type != "s3file":
        raise Exception(f"Invalid storage type: {storage_type}")

    context_from_arg["pipelineConfig"] = {
        **context_from_arg.get("pipelineConfig"),
        **resolve_secrets_in_pipeline_config(context_from_arg),
    }

    # override print function with our own, which decorates with workflow id and task id
    if should_override_builtin_print():
        __builtins__["print"] = construct_shadow_print(log)

    prefix = artifact_prefix if artifact_prefix else ""
    if not prefix and artifact_file_key:
        prefix = re.sub(r"ids/.*$", "", artifact_file_key)
    ids_util = create_ids_util(
        [
            {
                "bucket": artifact_bucket,
                "prefix": prefix,
                "endpoint": artifact_endpoint,
                "namespacePattern": r"^(common|client-.*)$",
            },
            {
                "name": "private",
                "bucket": artifact_bucket_private,
                "prefix": artifact_prefix_private if artifact_prefix_private else "",
                "endpoint": artifact_endpoint_private,
                "namespacePattern": r"^(private-.*)$",
            },
        ]
    )
    storage = Storage(storage_bucket, storage_file_key, storage_endpoint)
    datalake = Datalake(storage_endpoint, ids_util)
    func_config = FunctionConfig.from_task_script_dir(func_dir=func_dir, func_slug=func)
    context = Context(
        context_from_arg,
        datalake,
        ids_util,
        log,
        allowed_ids=func_config.allowed_ids,
    )
    func_module, func_name = resolve_func(func_dir, func)
    try:
        # TODO log at a finer level
        # log.log({ 'level': 'debug', 'tag': LOG_TAG_PRE_FUNCTION, 'funcName': f'{func_module}.{func_name}' })
        result = getattr(importlib.import_module(func_module), func_name)(
            input, context
        )
        # log.log({ 'level': 'debug', 'tag': LOG_TAG_POST_FUNCTION, 'funcName': func_name })

        if store_output:
            output_response(
                storage, {"type": COMPLETED, "result": result}, correlation_id
            )

        taskResult = {"status": "completed", "result": result or {}}
    except MemoryError as e:
        log.log(e)
        log.log({"level": "error", "message": "Got MemoryError"})
        os._exit(137)
    except FatalError as fatal_error:
        log.log(fatal_error)
        log.log(
            {"level": "error", "message": f"Got FatalError: '{fatal_error.message}'"}
        )
        os._exit(126)
    except Exception as e:
        log.log(e)

        if "SIGKILL" in str(e):
            log.log({"level": "error", "message": "Got SIGKILL - possible OOM"})
            os._exit(137)

        if store_output:
            output_response(storage, {"type": FAILED, "error": str(e)}, correlation_id)

        taskResult = {"status": "failed", "result": {"error": str(e)}}
    log.log({"level": "debug", "tag": LOG_TAG_SCRIPT_ENDED})
    return taskResult



from typing import List


def bytes_as_human_readable_string(num_bytes: float, decimals: int = 1) -> str:
    """Return the file size as a human-readable string."""
    for unit in ["", "Ki", "Mi", "Gi", "Ti"]:
        if abs(num_bytes) < 1024.0:
            return f"{round(num_bytes, ndigits=decimals):.{decimals}f} {unit}B"
        num_bytes /= 1024.0
    return f"{round(num_bytes, ndigits=decimals):.{decimals}f} YiB"


class UploadValidator:
    """
    Check whether given file(s) or other data may be uploaded to Tetra Data Platform.
    """

    def __init__(self, max_upload_size: int = 50 * 1024 * 1024) -> None:
        """
        :param max_upload_size:
            The maximum file size in bytes.
        """
        self.max_upload_size = max_upload_size

    def validate(self, upload_content: bytes) -> List[str]:
        """
        Check whether the provided data may be uploaded.

        :param upload_content:
            The data to be uploaded.
        :return:
            If the data is invalid, return a list containing the error message.
            Otherwise, return an empty list.
        """
        file_size = len(upload_content)
        if file_size > self.max_upload_size:
            friendly_file_size = bytes_as_human_readable_string(file_size)
            friendly_max_file_size = bytes_as_human_readable_string(
                self.max_upload_size
            )
            friendly_excess = bytes_as_human_readable_string(
                file_size - self.max_upload_size
            )
            return [
                f"File exceeded upload limit of ~{friendly_max_file_size} "
                f"by ~{friendly_excess}. Actual file size: {friendly_file_size}"
            ]

        return []



from functools import reduce

from .communication_format import (
    get_formats_up_to,
    get_communication_format,
    DEFAULT_COMMUNICATION_FORMAT,
    UNKNOWN_COMMUNICATION_FORMAT,
)


def __make_implementation_from_formats(history: dict, formats):
    return reduce(
        lambda implementation, communication_format: {
            **implementation,
            **history.get(communication_format, {}),
        },
        formats,
        {},
    )


def __make_unsupported_method(method_name: str):
    def method(*args, **kwargs):
        raise NotImplementedError(f"{method_name} is not implemented")

    return method


def __create_unsupported_format(implementation):
    return {
        key: __make_unsupported_method(key) if callable(value) else None
        for key, value in implementation.items()
    }


def __validate_history(history: dict):
    if DEFAULT_COMMUNICATION_FORMAT not in history:
        raise ValueError(
            f"{DEFAULT_COMMUNICATION_FORMAT} implementation is required in format history"
        )


def __decorate_history(history: dict) -> dict:
    default_unsupported_format = __create_unsupported_format(
        history[DEFAULT_COMMUNICATION_FORMAT]
    )
    unsupported_format = {
        **default_unsupported_format,
        **(
            history[UNKNOWN_COMMUNICATION_FORMAT]
            if UNKNOWN_COMMUNICATION_FORMAT in history
            else {}
        ),
    }
    return {**history, UNKNOWN_COMMUNICATION_FORMAT: unsupported_format}


def make_adapter(history: dict):
    """
    Allows you to declare an adapter through its format history, and a correct implementation
        will automatically be created at runtime
    :param history: A collection of partial implementations indexed by CommunicationFormat.
                    CommunicationFormat.V0 _must_ be supplied with a complete implementation
    :return: A merged implementation using methods up until the currently running CommunicationFormat
    """
    __validate_history(history)
    history = __decorate_history(history)
    current_format = get_communication_format()
    applicable_formats = get_formats_up_to(current_format)
    return Adapter(__make_implementation_from_formats(history, applicable_formats))


def select_versioned_value(history):
    """
    A bit of sugar for `make_adapter`
    Allows you to just supply values instead for format objects for each CommunicationFormat
    :param history:
    :return:
    """
    value = make_adapter(
        {format: {"value": value} for format, value in history.items()}
    ).value
    if value is None:
        raise NotImplementedError()
    return value


class Adapter(object):
    def __init__(self, implementation: dict):
        for k, v in implementation.items():
            setattr(self, k, v)



from urllib.parse import urlencode

import requests
import simplejson as json
from tenacity import retry, stop_after_attempt, wait_fixed, retry_if_exception_type

from ts_sdk.task.__util_adapters import (
    make_adapter,
    CommunicationFormat,
)
from ts_sdk.task.__util_adapters.endpoint_adapter import get_public_endpoint
from ts_sdk.task.encoders import DataclassEncoder

from ts_sdk.task import __util_ts_api as api

__RETRY_COUNT = 10
__RETRY_DELAY_SECONDS = 2


def get_endpoint():
    return get_public_endpoint("FILEINFO_ENDPOINT")


def request_format_adapter():
    return make_adapter(
        {
            CommunicationFormat.V0: {
                "get_headers": lambda org_slug: {"Content-Type": "application/json"},
                "get_labels_url": lambda org_slug, file_id: f"{get_endpoint()}/internal/{org_slug}/files/{file_id}/labels",
            },
            CommunicationFormat.V1: {
                "get_headers": lambda org_slug: {
                    "Content-Type": "application/json",
                    "x-org-slug": org_slug,
                },
                "get_labels_url": lambda org_slug, file_id: f"{get_endpoint()}/v1/fileinfo/files/{file_id}/labels",
            },
        }
    )


def format_pipeline_history_headers(context_data):
    # We set x-pipeline-id and x-pipeline-history to guard against self loops and circular pipelines
    return {
        "x-pipeline-id": context_data.get("pipelineId"),
        "x-workflow-id": context_data.get("workflowId"),
        "x-pipeline-history": pipeline_history_from_input_file_meta(context_data),
    }


retry_if_not_found = retry(
    wait=wait_fixed(__RETRY_DELAY_SECONDS),
    retry=retry_if_exception_type(FileNotFoundError),
    stop=stop_after_attempt(__RETRY_COUNT),
)


@retry_if_not_found
def add_labels(context_data, file_id, labels, no_propagate=False):
    org_slug = context_data.get("orgSlug")

    query_str = urlencode({"noPropagate": "true"} if no_propagate else {})
    url = f"{request_format_adapter().get_labels_url(org_slug, file_id)}?{query_str}"

    headers = {
        **request_format_adapter().get_headers(org_slug),
        **format_pipeline_history_headers(context_data),
    }
    response = requests.post(
        url,
        headers=headers,
        data=json.dumps(labels, cls=DataclassEncoder),
        verify=False,
    )

    if response.status_code == 200:
        print("Labels successfully added")
        return json.loads(response.text)
    elif response.status_code == 404:
        raise FileNotFoundError()
    else:
        print("Error adding labels: " + response.text)
        raise Exception(response.text)


@retry_if_not_found
def get_labels(context_data, file_id):
    org_slug = context_data.get("orgSlug")

    url = request_format_adapter().get_labels_url(org_slug, file_id)
    headers = request_format_adapter().get_headers(org_slug)
    response = requests.get(url, headers=headers, verify=False)

    if response.status_code == 200:
        print("Labels successfully obtained")
        return json.loads(response.text)
    elif response.status_code == 404:
        raise FileNotFoundError()
    else:
        print("Error getting labels: " + response.text)
        raise Exception(response.text)


@retry_if_not_found
def delete_labels(context_data, file_id, label_ids):
    org_slug = context_data.get("orgSlug")

    suffix = "&".join(map(lambda id: "id=" + str(id), label_ids))
    url = request_format_adapter().get_labels_url(org_slug, file_id)
    if suffix:
        url += f"?{suffix}"
    pipeline_id = context_data.get("pipelineId")

    # We set x-pipeline-id and x-pipeline-history to guard against self loops and circular pipelines
    headers = {
        **request_format_adapter().get_headers(org_slug),
        **format_pipeline_history_headers(context_data),
    }
    response = requests.delete(url, headers=headers, verify=False)

    if response.status_code == 200:
        print("Labels successfully deleted")
        return json.loads(response.text)
    elif response.status_code == 404:
        raise FileNotFoundError()
    else:
        print("Error deleting labels: " + response.text)
        raise Exception(response.text)


@retry_if_not_found
def get_file_pointer(context_data, file_id):
    org_slug = context_data.get("orgSlug")
    response = api.get(
        f"/v1/fileinfo/file/{file_id}",
        headers=request_format_adapter().get_headers(org_slug),
    )
    if response.status_code == 200:
        file_info = json.loads(response.text)
        file = file_info["file"]
        return {
            "type": "s3file",
            "fileId": file_id,
            "bucket": file["bucket"],
            "fileKey": file["path"],
            "version": file["version"],
        }
    elif response.status_code == 404:
        raise FileNotFoundError(f"File with fileId {file_id} not found")
    else:
        raise Exception(response.text)


# When we write a new file or update metadata or tags, we add the pipeline id and pipeline history to the
# new file's s3 metadata fields. If we only add labels, we don't update the file's s3 metadata fields because
# we don't generate a new file. To keep track of a file's pipeline history, we will have to use the
# `inputFile.meta` object that contains information about a file that may not exist in the s3 metadata fields.
def pipeline_history_from_input_file_meta(context_data):
    pipeline_id = context_data.get("pipelineId")
    if "pipelineHistory" in context_data.get("inputFile", {}).get("meta", {}):
        existing_pipeline_history = context_data["inputFile"]["meta"]["pipelineHistory"]
        # If the pipeline history field exists, but it is an empty string, we don't want a leading comma
        if len(existing_pipeline_history) == 0:
            new_pipeline_history = pipeline_id
        else:
            # Only add the pipeline's id if it is not already present in the history
            if pipeline_id in existing_pipeline_history:
                new_pipeline_history = existing_pipeline_history
            else:
                new_pipeline_history = existing_pipeline_history + "," + pipeline_id
    else:
        new_pipeline_history = pipeline_id
    return new_pipeline_history



# These should be kept in sync with S3MetadataFields.js in ts-lib-shared-schema
# https://github.com/tetrascience/ts-lib-shared-schema/blob/master/src/S3MetadataFields.js
FIELDS = {
    "VERSION": "ts_version",
    "INTEGRATION_TYPE": "ts_integration_type",
    "INTEGRATION_NAME": "ts_integration_name",
    "INTEGRATION_ID": "ts_integration_id",  # TODO new?
    "FILE_ID": "ts_file_id",
    # FILE_TYPE is deprecated. This field is no longer used, it could contain not valid file type
    #  as some sources were not setting it correctly
    "FILE_TYPE": "ts_processed_file_type",
    "FILE_PATH": "ts_file_path",  # how to specify this?
    "FILE_NAME": "ts_file_name",
    "SOURCE_ID": "ts_integration_source",
    "SOURCE_TYPE": "ts_source_type",
    "SOURCE_NAME": "ts_source_name",
    "DESTINATION_ID": "ts_destination_id",
    "RAW_FILE_ID": "ts_source_file_id",
    "RAW_FILE_VERSION": "ts_raw_file_version",
    "IDS": "ts_ids",
    "IDS_NAMESPACE": "ts_ids_namespace",
    "IDS_TYPE": "ts_ids_type",
    "IDS_VERSION": "ts_ids_type_version",
    "CUSTOM_METADATA": "ts_integration_metadata",
    "CUSTOM_TAGS": "ts_integration_tags",
    "PIPELINE_ID": "ts_pipeline_id",
    "PIPELINE_MASTER_SCRIPT": "ts_master_script",
    "PIPELINE_TASK_SCRIPT": "ts_task_script",
    "PIPELINE_TASK_SLUG": "ts_task_slug",
    "PIPELINE_TASK_EXECUTION_ID": "ts_task_execution_id",
    "PIPELINE_WORKFLOW_ID": "ts_workflow_id",
    "PIPELINE_HISTORY": "ts_pipeline_history",
    "TRACE_ID": "ts_trace_id",
    "CONTENT_CREATED_FROM_FILE_ID": "ts_content_created_from_file_id",
    "DO_NOT_INHERIT_LABELS": "ts_do_not_inherit_labels",
    "API_USER_ID": "ts_user_id",
}



from itertools import chain
from typing import Any, Dict, Iterable, Set, Tuple, Type

from ts_sdk.task.log_codes.log_code import LogCode
from ts_sdk.task.log_codes.log_code_validator import LogCodeValidator


class LogCodeConfigurationError(Exception):
    """Raised when there is an error defining a log code."""


class LogCodeCollectionMeta(type):
    """
    Metaclass for singleton collection of Log Codes.

    This metaclass validates that all Log Code members are valid using a LogCodeValidator and
    provides convenience functions for working with Log Codes. See the Log Code Guide document
    for further information.
    """

    def __new__(
        mcs, name: str, bases: Tuple[Type, ...], namespace: Dict[str, Any], **_kwargs
    ) -> "LogCodeCollectionMeta":
        """
        :param name:
            The name of the class returned by this method.
        :param bases:
            The base classes of the class.
        :param namespace:
            A mapping of the class' member names to their values.
        """
        current_class_log_codes = mcs._get_tetrascience_log_codes(namespace.values())
        base_log_codes: Iterable[LogCode] = chain.from_iterable(
            mcs._get_tetrascience_log_codes(base_class.__dict__.values())
            for base_class in bases
        )

        common_log_codes = mcs.get_common_log_codes(
            current_class_log_codes, base_log_codes
        )
        if len(common_log_codes) > 0:
            raise LogCodeConfigurationError(
                "Cannot override parent members in child classes. "
                f"The base class and the {name} class have the following pairs of log "
                f"codes with common messages or integer codes: "
                f"{', '.join(map(str, common_log_codes))}. Please change the message or "
                f"integer code in the {name} class."
            )

        validator = LogCodeValidator()
        invalid_log_codes = {
            log_code
            for log_code in current_class_log_codes
            if validator.get_validation_errors(log_code)
        }
        if len(invalid_log_codes) > 0:
            raise LogCodeConfigurationError(
                f"The following TetraScience log codes are invalid: "
                f"{', '.join(map(str, sorted(invalid_log_codes)))}. Ensure that "
                f"the log code is in the correct range, the log code message is in PascalCase "
                f"and is less than {validator.MAX_CODE_MESSAGE_LENGTH} characters long."
            )

        return super(LogCodeCollectionMeta, mcs).__new__(
            mcs, name, bases, namespace, **_kwargs
        )

    @staticmethod
    def get_common_log_codes(
        log_codes: Iterable[LogCode], base_log_codes: Iterable[LogCode]
    ) -> Set[Tuple[LogCode, LogCode]]:
        """Return pairs of log codes that have the same message or ID."""
        log_codes = set(log_codes)
        base_log_codes = set(base_log_codes)

        tetrascience_ids = {log_code.code: log_code for log_code in log_codes}
        base_codes = {log_code.code: log_code for log_code in base_log_codes}
        common_codes = set(tetrascience_ids).intersection(base_codes)

        tetrascience_code_messages = {
            log_code.code_message: log_code for log_code in log_codes
        }
        base_code_messages = {
            log_code.code_message: log_code for log_code in base_log_codes
        }
        common_code_messages = set(tetrascience_code_messages).intersection(
            base_code_messages
        )

        return {
            *(
                (tetrascience_ids[log_code], base_codes[log_code])
                for log_code in common_codes
            ),
            *(
                (tetrascience_code_messages[message], base_code_messages[message])
                for message in common_code_messages
            ),
        }

    @staticmethod
    def _get_tetrascience_log_codes(members: Iterable[Any]) -> Set[LogCode]:
        """
        Return all the unique values that are TetraScience log codes. A log code is
        a TetrasScience log code iff its ID is in [0, 5000).
        """
        all_log_codes: Set[LogCode] = {
            member for member in members if isinstance(member, LogCode)
        }
        return {member for member in all_log_codes if 0 <= member.code < 5000}



import argparse
import base64
import json
import multiprocessing
import os
import shutil
import sys
import tempfile
import traceback
from .__util_log import Log
from .__util_task import update_task_status
from .__task_script_runner import run

log = Log({})


def get_run_params(task):
    return {
        "input": task.get("input"),
        "context_from_arg": task.get("context"),
        "func": task.get("func"),
        "correlation_id": task.get("correlation_id"),
        "func_dir": task.get("func_dir"),
        "store_output": False,
        "storage_type": os.environ.get("TASK_STORAGE_TYPE"),
        "storage_bucket": os.environ.get("TASK_STORAGE_S3FILE_BUCKET"),
        "storage_file_key": os.environ.get("TASK_STORAGE_S3FILE_FILE_KEY"),
        "storage_endpoint": os.environ.get("TASK_STORAGE_S3FILE_ENDPOINT"),
        "artifact_bucket": os.environ.get("ARTIFACT_S3FILE_BUCKET"),
        "artifact_prefix": os.environ.get("ARTIFACT_S3FILE_PREFIX"),
        "artifact_endpoint": os.environ.get("ARTIFACT_S3FILE_ENDPOINT"),
        "artifact_file_key": os.environ.get("ARTIFACT_IDS_SCHEMA_S3FILE_FILE_KEY"),
        "artifact_bucket_private": os.environ.get("ARTIFACT_S3FILE_BUCKET_PRIVATE"),
        "artifact_prefix_private": os.environ.get("ARTIFACT_S3FILE_PREFIX_PRIVATE"),
        "artifact_endpoint_private": os.environ.get("ARTIFACT_S3FILE_ENDPOINT_PRIVATE"),
    }


def task_process_fn(task, shared_dict):
    task_tmp_dir = tempfile.mkdtemp()
    os.environ.update({"TMPDIR": task_tmp_dir})

    run_params = get_run_params(task)
    sys.path.append(run_params.get("func_dir"))
    try:
        shared_dict["result"] = run(**run_params)
    except:
        e = sys.exc_info()[1]
        log.log(log.generate_error(e))
        shared_dict["error"] = traceback.format_exc()
    finally:
        sys.path.remove(run_params.get("func_dir"))
        shutil.rmtree(task_tmp_dir, ignore_errors=True)


def run_task_process(task, task_process_cb=None):
    manager = multiprocessing.Manager()
    shared_dict = manager.dict({"result": None, "error": None})
    task_id = task.get("id")
    task_process = multiprocessing.Process(
        name=f"task-{task_id}", target=task_process_fn, args=(task, shared_dict)
    )
    if task_process_cb:
        task_process_cb(task_process)
    task_process.start()
    task_process.join()
    exitcode = task_process.exitcode
    if exitcode != 0:
        if exitcode == -9 or exitcode == 137:
            exitcode = 137
            update_status_payload = {
                "status": "failed",
                "exitCode": exitcode,
                "result": {
                    "error": {
                        "message": {
                            "text": f"Invalid exit code {exitcode}",
                            "oomError": True,
                        }
                    }
                },
            }
        else:
            update_status_payload = {
                "status": "failed",
                "exitCode": exitcode,
                "result": {"error": f"Invalid exit code {exitcode}"},
            }
    else:
        if not (shared_dict["result"] is None):
            update_status_payload = {"exitCode": 0, **shared_dict["result"]}
        else:
            update_status_payload = {
                "status": "failed",
                "exitCode": 0,
                "result": {
                    "error": (
                        shared_dict["error"]
                        if shared_dict["error"]
                        else "No content returned by worker"
                    )
                },
            }
    update_task_status(task, update_status_payload)
    return shared_dict


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--func", help="function slug", required=True)
    parser.add_argument("--correlation-id", help="correlation id")
    parser.add_argument("--input", help="input string", default="")
    parser.add_argument("--context", help="context", default="")
    parser.add_argument("--func-dir", help="function dir", default="./func")
    args = parser.parse_args()
    params = {
        "input": json.loads(base64.standard_b64decode(args.input)),
        "context_from_arg": json.loads(base64.standard_b64decode(args.context)),
        "func": args.func,
        "correlation_id": args.correlation_id,
        "func_dir": args.func_dir,
        "storage_type": os.environ.get("TASK_STORAGE_TYPE"),
        "storage_bucket": os.environ.get("TASK_STORAGE_S3FILE_BUCKET"),
        "storage_file_key": os.environ.get("TASK_STORAGE_S3FILE_FILE_KEY"),
        "storage_endpoint": os.environ.get("TASK_STORAGE_S3FILE_ENDPOINT"),
        "artifact_bucket": os.environ.get("ARTIFACT_S3FILE_BUCKET"),
        "artifact_prefix": os.environ.get("ARTIFACT_S3FILE_PREFIX"),
        "artifact_endpoint": os.environ.get("ARTIFACT_S3FILE_ENDPOINT"),
        "artifact_file_key": os.environ.get("ARTIFACT_IDS_SCHEMA_S3FILE_FILE_KEY"),
        "artifact_bucket_private": os.environ.get("ARTIFACT_S3FILE_BUCKET_PRIVATE"),
        "artifact_prefix_private": os.environ.get("ARTIFACT_S3FILE_PREFIX_PRIVATE"),
        "artifact_endpoint_private": os.environ.get("ARTIFACT_S3FILE_ENDPOINT_PRIVATE"),
    }
    sys.path.append(params["func_dir"])

    run(**params)



import json
import typing as t
from importlib.resources import read_text


def load(name: str) -> t.Any:
    return json.loads(read_text(__package__, name))


config = load("config.schema.json")

protocol = load("protocol.schema.json")
protocolV3 = load("protocol.v3.schema.json")

__all__ = ["config", "protocol", "protocolV3"]



import functools


def deprecated(message):
    def deprecated_decorator(func):
        @functools.wraps(func)
        def deprecated_func(*args, **kwargs):
            print(
                {
                    "level": "warn",
                    "message": message,
                }
            )
            return func(*args, **kwargs)

        return deprecated_func

    return deprecated_decorator


def return_on_failure(errors=(Exception,), default_value=None):
    def decorator(func):
        @functools.wraps(func)
        def applicator(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except errors:
                return default_value

        return applicator

    return decorator


def raise_on_failure(errors=(Exception,), exception=Exception):
    def decorator(func):
        @functools.wraps(func)
        def applicator(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except errors:
                raise exception

        return applicator

    return decorator


def cached(function):
    """
    Decorates a function such that is only executes a single time
    The return value is cached and returned for every future invocation
    IMPORTANT: Will not re-execute even if arguments change
    :param function: the method to be decorated
    :return: a function with the same signature as the provided function
    """
    result = None

    @functools.wraps(function)
    def applicator(*args, **kwargs):
        nonlocal result
        if kwargs.get("force"):
            result = None
            kwargs.pop("force")
        if result is None:
            result = function(*args, **kwargs)
        return result

    return applicator


def predicated_on(predicate, message):
    def decorator(function):
        @functools.wraps(function)
        def applicator(*args, **kwargs):
            if predicate(*args, **kwargs) is False:
                raise ValueError(message)
            return function(*args, **kwargs)

        return applicator

    return decorator



import urllib.parse

from ts_sdk.task.__util_adapters import (
    CommunicationFormat,
    select_versioned_value,
)

import os
import traceback

import requests
from requests.exceptions import RequestException

from ts_sdk.task.__util_adapters.endpoint_adapter import get_public_endpoint

task_group_hash = os.environ.get("TASK_GROUP_HASH")
container_id = os.environ.get("CONTAINER_ID")


def get_search_eql_api_url():
    return get_public_endpoint("ORCHESTRATOR_ENDPOINT") + select_versioned_value(
        {
            CommunicationFormat.V0: "/datalake/searchEql",
            CommunicationFormat.V1: "/v1/datalake/searchEql",
        }
    )


def es_datalake_search_eql(payload, query=None):
    try:
        url_encoded_query = urllib.parse.urlencode(query or {}, doseq=True)
        url_suffix = f"?{url_encoded_query}" if url_encoded_query else ""
        search_url = get_search_eql_api_url() + url_suffix
        response = requests.post(
            search_url,
            json=payload,
            headers={
                "x-task-group-hash": task_group_hash,
                "x-container-id": container_id,
                "content-type": "application/json; charset=utf-8",
            },
            verify=False,
        )
        if response.status_code >= 400:
            print({"level": "error", "message": response.text})
            raise Exception(f"Got {response.status_code} for {search_url}")
        return response.json()
    except RequestException:
        print({"level": "error", "message": traceback.format_exc()})
        raise


def es_hit_to_file_pointer(hit):
    es_file = hit["_source"]["file"]
    return {
        "type": "s3file",
        "bucket": es_file["bucket"],
        "fileKey": es_file["path"],
        "version": es_file["version"],
        "fileId": hit["_source"]["fileId"],
    }



import os
import pathlib
import time
import xml.etree.ElementTree as ET
import zipfile
from pathlib import Path
from typing import Iterable, Optional, Union

import requests
from colorama import Fore


def get_latest_version():
    try:
        r = requests.get("https://pypi.org/rss/project/ts-sdk/releases.xml", timeout=5)
        root = ET.fromstring(r.content)
        return root.findall("channel/item/title")[0].text
    except:
        return "0.0.0"


def check_update_required(current_version):
    try:
        latest_version_path = Path.home() / ".ts-sdk.latest"
        latest_version = "0.0.0"

        # refresh saved latest version once per day
        if (
            latest_version_path.is_file()
            and time.time() - latest_version_path.stat().st_mtime < 24 * 3600
        ):
            latest_version = latest_version_path.read_text()
        else:
            latest_version = get_latest_version()
            latest_version_path.write_text(latest_version)

        if latest_version and check_versions_for_update(
            current_version, latest_version
        ):
            print(
                f"\n{Fore.YELLOW}Please upgrade ts-sdk (local: {current_version}, latest: {latest_version}){Fore.RESET}"
            )
            print(f"{Fore.YELLOW}Use: pip3 install ts-sdk --upgrade{Fore.RESET}\n")

    except Exception as ex:
        # print(ex)
        pass


def check_versions_for_update(current: str, latest: str):
    current_major, current_minor, *rest = current.split(".")
    latest_major, latest_minor, *rest = latest.split(".")
    if int(current_major) < int(latest_major):
        return True
    if int(current_minor) < int(latest_minor):
        return True
    return False


def zipdir(
    path: Union[str, pathlib.Path],
    zip_archive: zipfile.ZipFile,
    folders_to_exclude: Optional[Iterable[str]] = None,
) -> None:
    if folders_to_exclude is None:
        folders_to_exclude = set()
    folders_to_exclude = set(folders_to_exclude)
    for root, folders, files in os.walk(path, topdown=True):
        actually_excluded_folders = set(folders).intersection(folders_to_exclude)
        for folder in actually_excluded_folders:
            folders.remove(folder)
        for file in files:
            local_path = os.path.join(root, file)
            zip_path = os.path.relpath(os.path.join(root, file), os.path.join(path))
            zip_archive.write(local_path, zip_path)



import json
import os
import sys
from importlib import import_module

import jsonschema
import yaml
from pytest import fixture

import ts_sdk.schemas as s


@fixture
def config():
    with open("./config.json") as f:
        return json.load(f)


@fixture
def protocol():
    with open("../protocol/protocol.yml") as f:
        return yaml.safe_load(f)


def test_config_schema(config):
    jsonschema.validate(config, s.config)


def test_funcs(config):
    for fn in config["functions"]:
        qname = fn["function"]
        mod, func = qname.split(".")
        assert hasattr(import_module(mod), func), f"unable to find function {qname}"
        assert func == fn["slug"]


def test_protocol_schema(protocol):
    jsonschema.validate(protocol, s.protocolV3)



import re
from typing import Any, Collection, List, Optional

from ts_sdk.task.data_model import AnyLabel, Label

meta_name_reg = re.compile(r"^[0-9a-zA-Z-_+ ]+$")
meta_value_reg = re.compile(r"^[0-9a-zA-Z-_+.,/ ]+$")
tag_reg = re.compile(r"^[0-9a-zA-Z-_+./ ]+$")
label_reg = re.compile(r"[^\uFFFE\uFFFF\n\r\\]+")
string_is_trimmed_reg = re.compile(r"^\S(.*\S)?$")

LABEL_NAME_MAX_CHARACTER_LIMIT = 128
LABEL_VALUE_MAX_CHARACTER_LIMIT = 256
TAG_MAX_CHAR_LIMIT = 128


def validate_file_meta(meta) -> bool:
    if meta is None:
        return True
    for key, value in meta.items():
        if not meta_name_reg.match(key):
            raise ValueError(
                f"Invalid metadata key {key}! Expected pattern: {meta_name_reg.pattern}"
            )
        if not meta_value_reg.match(str(value)):
            raise ValueError(
                f"Invalid metadata value {value}! Expected pattern: {meta_value_reg.pattern}"
            )
        validate_string_is_trimmed(key)
        validate_string_is_trimmed(value)
    return True


def validate_file_tags(tags: List[str]) -> bool:
    """
    Validate file tags as per https://developers.tetrascience.com/docs/basic-concepts-metadata-tags-and-labels
    """
    if tags is None:
        return True
    for tag in tags:
        if not isinstance(tag, str):
            raise TypeError(
                f"Invalid Tag. Expected type(tag) is str. Received type: {type(tag)}. Tag: {tag}"
            )
        if len(tag) > TAG_MAX_CHAR_LIMIT:
            raise ValueError(f"Tag length is greater that 128 chars. tag: {tag}")
        if not tag_reg.match(str(tag)):
            raise ValueError(f"Invalid tag {tag}! Expected pattern: {tag_reg.pattern}")
    return True


def cast_iterable(maybe_iterable, identifier) -> Collection:
    try:
        return iter(maybe_iterable)
    except TypeError:
        raise ValueError(
            f"{identifier} must be an iterable.  Given {type(maybe_iterable)}."
        )


def validate_not_none(maybe_none, identifier) -> bool:
    if maybe_none is None:
        raise ValueError(f"{identifier} should not be None.  Given None.")
    return True


def validate_not_empty(maybe_empty, identifier) -> bool:
    collection = cast_iterable(maybe_empty, identifier)
    if len(collection) == 0:
        raise ValueError(
            f"{identifier} must be a non-empty iterable of labels. Given an empty iterable."
        )
    return True


def validate_file_labels(labels: Optional[Collection[AnyLabel]]) -> bool:
    """
    Validates the list of labels given.
    List of labels must be made up of Label dictionaries
    """
    if labels is None or len(labels) == 0:
        return False

    # make sure the input is iterable
    labels_iter = cast_iterable(labels, "Labels")
    if labels_iter:
        valid_keys = {"name", "value"}
        errors = []
        for label in labels_iter:
            if not isinstance(label, (dict, Label)):
                errors.append(
                    f'Label must be a dictionary of the form {{"name": "foo", "value": "bar"}} or an instance of "ts_sdk.task.types.Label". Given: {label} ({type(label)})'
                )
                continue
            if isinstance(label, dict):
                # validate the shape of the dictionary
                label_keys = set(label.keys())
                if valid_keys - label_keys:
                    errors.append(
                        f"Label is missing required keys: {list(valid_keys - label_keys)}. Given: {label}"
                    )
                    continue

                if label_keys - valid_keys:
                    errors.append(
                        f"Label has extra keys: {list(label_keys - valid_keys)}. Given: {label}"
                    )
                    continue

                # validate the types of the inputs
                label_name = label["name"]
                label_value = label["value"]
            else:
                # no need to validate the shape of a dataclass
                label_name = label.name
                label_value = label.value

            if not isinstance(label_name, str):
                errors.append(
                    f'Invalid label name "{label_name}"! Must be a string. Given {type(label_name)}'
                )
                continue
            if not isinstance(label_value, str):
                errors.append(
                    f'Invalid label value "{label_value}"! Must be a string. Given {type(label_value)}'
                )
                continue

            # validate the characters used in the inputs
            if label_reg.fullmatch(label_name) is None:
                errors.append(
                    f'Invalid label name "{label_name}"! Expected pattern: "{label_reg.pattern}"'
                )
                continue
            if label_reg.fullmatch(label_value) is None:
                errors.append(
                    f'Invalid label value "{label_value}"! Expected pattern: "{label_reg.pattern}"'
                )
                continue

            # Validate character limits
            if len(label_name) > LABEL_NAME_MAX_CHARACTER_LIMIT:
                errors.append(
                    f"Invalid label_name length. Max limit is {LABEL_NAME_MAX_CHARACTER_LIMIT} characters"
                    f"label name: {label_name}"
                )
                continue

            if len(str(label_value)) > LABEL_VALUE_MAX_CHARACTER_LIMIT:
                errors.append(
                    f"Invalid label_value length. Max limit is {LABEL_VALUE_MAX_CHARACTER_LIMIT} characters"
                    f"label value: {label_value}"
                )
                continue
        if errors:
            error_msg = "\n".join(errors)
            raise ValueError(f"Invalid label(s).  Errors:\n{error_msg}")
    return True


def validate_string_is_trimmed(value: Any) -> bool:
    """Validates that the given value has no leading or trailing whitespace

    If the given value is not a string, its `str()` representation is checked instead.
    """
    # Need to typecast non-string values to string in order to use the regex
    string_value = str(value)
    if not string_is_trimmed_reg.match(string_value):
        raise ValueError(f'Value "{value}" cannot contain leading or trailing spaces!')
    return True



import functools

from colorama import Fore

MESSAGE = """
This cli is being deprecated,
and will no longer be supported starting July 31st, 2025.
Migrate to the new TetraScience cli by running
pip3 install tetrascience-cli
"""


def deprecated(func):
    @functools.wraps(func)
    def deprecated_func(*args, **kwargs):
        print(f"{Fore.YELLOW}{MESSAGE}{Fore.RESET}")
        return func(*args, **kwargs)

    return deprecated_func



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Optional


@dataclass
class VersionedRef:
    composite: Optional[str] = None
    namespace: Optional[str] = None
    name: Optional[str] = None
    version: Optional[str] = None

    def __init__(self, composite=None, namespace=None, name=None, version=None):
        """Converts this VersionedRef into canonical form

        After a call to this function, the composite field matches the
        other three fields. The function raises an exception if the
        field values are inconsistent.
        """

        if composite is None:
            self.namespace, self.name, self.version = namespace, name, version
            self.composite = f"{namespace}/{name}:{version}"
            return

        m = re.fullmatch("(.*)/(.*):(.*)", composite)
        if m is None:
            raise ValueError("versioned ref doesn't match namespace/name:version")
        self.composite = composite
        self.namespace, self.name, self.version = m.groups()

        if namespace is not None and namespace != self.namespace:
            raise ValueError("inconsistent namespace")
        if name is not None and name != self.name:
            raise ValueError("inconsistent name")
        if version is not None and version != self.version:
            raise ValueError("inconsistent version")



import os
import shutil
from argparse import ArgumentParser
from pathlib import Path

from ts_sdk.cli.__deprecated import deprecated


def init_cmd_args(parser: ArgumentParser):
    parser.add_argument(
        "--protocol_slug", "-p", type=str, required=True, help="Slug of the protocol"
    )
    parser.add_argument(
        "--task_script_slug",
        "-t",
        type=str,
        required=True,
        help="Slug of the task script",
    )
    parser.add_argument(
        "--folder_path", "-f", type=str, required=True, help="Destination folder"
    )
    parser.add_argument(
        "--remove",
        "-r",
        action="store_true",
        help="Force removal/overwrite if folder already exists",
    )
    parser.add_argument(
        "--org",
        "-o",
        type=str,
        required=True,
        help="Org into which the scripts will be uploaded",
    )
    parser.add_argument(
        "--preserve_templates",
        type=bool,
        default=False,
        help="If true, leave template files on disk instead of deleting them",
    )
    parser.add_argument(
        "--protocol_schema",
        type=str,
        default="v3",
        required=False,
        choices=["v2", "v3"],
        help="If set to 'v2', uses legacy protocol template",
    )
    parser.set_defaults(func=__cmd)


def copy_from_root_to_root(src_root, dst_root):
    def copy(src_relative, dst_relative):
        src_path = os.path.join(src_root, src_relative)
        dst_path = os.path.join(dst_root, dst_relative)
        shutil.copytree(src_path, dst_path)

    return copy


@deprecated
def __cmd(args):
    # Remove folder if it exists
    new_folder_path = os.path.abspath(args.folder_path)
    new_folder = Path(new_folder_path)
    if new_folder.exists():
        print(f"{new_folder_path} already exists")
        if args.remove:
            shutil.rmtree(new_folder_path)
            print(f"{new_folder_path} removed")
        else:
            print(
                f"{new_folder_path} exists, do nothing. You can use --remove to remove/overwrite"
            )
            return
    copy_from_template_to_new_folder = copy_from_root_to_root(
        src_root=os.path.join(os.path.dirname(__file__), "protocol-template"),
        dst_root=new_folder_path,
    )
    copy_from_template_to_new_folder(
        src_relative="task-script", dst_relative="task-script"
    )
    copy_from_template_to_new_folder(
        src_relative=("protocol-v2" if args.protocol_schema == "v2" else "protocol"),
        dst_relative="protocol",
    )
    for p in Path(new_folder_path).glob("**/*.template"):
        text = p.read_text()
        text = text.replace("{{ org }}", args.org)
        text = text.replace("{{ protocol_slug }}", args.protocol_slug)
        text = text.replace("{{ task_script_slug }}", args.task_script_slug)

        dest = p.with_name(p.stem)
        print(f"Generating {dest}")
        dest.write_text(text)

        if not args.preserve_templates:
            print(f"Deleting {p}")
            p.unlink()



from functools import partial


def curry_for_each_method(format_implementation: dict, *args, **kwargs):
    """
    A _very_ naive implementation of currying, for each callable value inside a dict
    Created to support the common use of currying URLs for API adapters
    Only curries to level callables -- does not recur down a dict tree
    :param format_implementation: The dict that may contain callables
    :param args:
    :param kwargs:
    :return: the dict containing curried functions
    """
    return {
        key: partial(value, *args, **kwargs) if callable(value) else value
        for key, value in format_implementation.items()
    }



import re
from typing import List

from ts_sdk.task.log_codes.log_code import LogCode


class LogCodeValidator:
    """Validator for Log Codes

    Checks code is in valid range, message is alphanumeric and in PascalCase, and message is
    correct length"""

    #: The maximum number of characters a Log Code message can have.
    MAX_CODE_MESSAGE_LENGTH = 32
    #: Regular expression that the Log Code message must match (PascalCase).
    CODE_MESSAGE_REGEX = re.compile(r"^[A-Z\d][a-z\d]+(?:[A-Z\d][a-z\d]+)*$")

    @classmethod
    def get_validation_errors(cls, log_code: LogCode) -> List[str]:
        """
        Return validation errors for given Log Code.

        :param log_code:
            The TetraScience log code to validate. A TetraScience log code is a log code whose
            ID is >= 1000 and < 10000. Log codes > 0 and < 1000 are still valid, but for internal
            use only.
        :return:
            List of reasons why the code is invalid, or empty list if code is valid.
        """
        errors = []
        if log_code.code < 1 or log_code.code > 9999:
            errors.append(
                f"Log Code outside of valid range [0, 9999]. Code: {log_code.code}"
            )
        if log_code.code_message.isalnum():
            if (
                0 < log_code.code < 5000
                and cls.CODE_MESSAGE_REGEX.match(log_code.code_message) is None
            ):
                errors.append(
                    f'Log Code message is not PascalCase. Message: "{log_code.code_message}"'
                )
        else:
            errors.append(
                f"Log Code message contains non-alphanumeric characters. Message: "
                f'"{log_code.code_message}"'
            )
        if len(log_code.code_message) > cls.MAX_CODE_MESSAGE_LENGTH:
            errors.append(
                f"Log Code message longer than {cls.MAX_CODE_MESSAGE_LENGTH} characters"
            )
        return errors



import datetime
import os
import traceback
from enum import Enum
from os.path import relpath
from typing import Any, Dict, Mapping, Optional

import simplejson as json

from ts_sdk.task.log_codes.log_code import LogCode
from ts_sdk.task.log_codes.log_code_collection import LogCodeCollection
from ts_sdk.task.log_codes.log_code_validator import LogCodeValidator

TIMESTAMP_KEY = "timestamp"
ORG_SLUG_KEY = "orgSlug"
MESSAGE_KEY = "message"
FILE_ID_KEY = "fileId"
MASTER_SCRIPT_KEY = "masterScript"
TASK_SCRIPT_KEY = "taskScript"
TRACE_ID_KEY = "traceId"
TASK_ID_KEY = "taskId"
WORKFLOW_ID_KEY = "workflowId"
CONTAINER_ID_KEY = "containerId"
LEVEL_KEY = "level"
ERROR_TYPE_KEY = "errorType"
EXCEPTION_KEY = "exception"


class LogLevel(str, Enum):
    """Log Level strings"""

    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"


class Log:
    """Logging class for ts-sdk-python"""

    def __init__(self, context):
        self._context = context

    def log(self, *args, sep=" "):
        """Logs a JSON object created from function arguments"""
        try:
            # Perform actual logging, that will be transported to Cloudwatch.
            __builtins__["default_print"](
                json.dumps(self.create_json_log_entry(args, sep=sep))
            )
        except TypeError as ex:
            __builtins__["default_print"](json.dumps(self.create_json_log_entry(ex)))
            for arg in args:
                __builtins__["default_print"](arg)

    def create_json_log_entry(self, input_, sep=" "):
        """Create JSON object for logging"""
        log_entry = self.generate_default()

        if isinstance(input_, tuple):
            if len(input_) > 1:
                log_entry = {
                    **log_entry,
                    MESSAGE_KEY: sep.join(str(item) for item in input_),
                    LEVEL_KEY: LogLevel.INFO,
                }
                return log_entry
            if len(input_) == 1:
                input_ = input_[0]

        if isinstance(input_, str):
            log_entry = {**log_entry, MESSAGE_KEY: input_, LEVEL_KEY: LogLevel.INFO}
        elif isinstance(input_, Exception):
            log_entry = {**log_entry, **self.generate_error(input_)}
        elif isinstance(input_, Mapping):
            log_entry = {**log_entry, **input_}
        else:
            log_entry = {
                **log_entry,
                MESSAGE_KEY: str(input_),
                LEVEL_KEY: LogLevel.INFO,
            }
        if LEVEL_KEY not in log_entry:
            log_entry[LEVEL_KEY] = LogLevel.INFO

        return log_entry

    def generate_default(self):
        """Generate default log fields"""
        try:
            default_log = {
                TIMESTAMP_KEY: datetime.datetime.now().isoformat(),
            }

            if "CONTAINER_ID" in os.environ:
                default_log[CONTAINER_ID_KEY] = os.environ.get("CONTAINER_ID")

            if "traceId" in self._context.get("inputFile", {}).get("meta", {}):
                default_log[TRACE_ID_KEY] = self._context["inputFile"]["meta"][
                    "traceId"
                ]

            if "taskId" in self._context:
                default_log[TASK_ID_KEY] = self._context.get("taskId")

            if "workflowId" in self._context:
                default_log[WORKFLOW_ID_KEY] = self._context.get("workflowId")
        except (AttributeError, KeyError) as exc:
            __builtins__["default_print"]("Error in default log")
            __builtins__["default_print"](exc)
            default_log = {}

        return default_log

    def generate_error(self, err):
        """Generate default error message"""
        tb_lines = traceback.format_exception(err.__class__, err, err.__traceback__)
        tb_text = "".join(tb_lines)
        error = {
            LEVEL_KEY: LogLevel.ERROR,
            ERROR_TYPE_KEY: str(type(err)),
            EXCEPTION_KEY: tb_text,
        }
        if hasattr(err, "message"):
            error[MESSAGE_KEY] = err.message
        else:
            error[MESSAGE_KEY] = str(err)

        return error

    def _log_with_level(
        self,
        message: str,
        log_level: LogLevel,
        code: Optional[LogCode] = None,
        *,
        extra: Optional[Mapping[str, Any]] = None,
    ):
        """Log a message with chosen log level

        Logs an error in addition to given message if:
          * LogCode is invalid
          * Log level is ERROR, WARNING but LogCode is None
          * Message passed is not a <str>
        """
        errors = []
        original_log = {}
        fields = {"level": log_level}

        if not isinstance(message, str):
            errors.append(
                f"Expected message to be 'str', received {type(message)} instead"
            )
            message = str(message)
            original_log["message"] = message
        fields["message"] = message

        if code is not None:
            if not isinstance(code, LogCode):
                errors.append(f"Expected {type(LogCode)}, got {type(code)} instead")
                code = LogCodeCollection.generic
                original_log["code"] = code
            else:
                validator_errors = LogCodeValidator.get_validation_errors(code)
                if validator_errors:
                    errors.extend(validator_errors)
                    original_log["code"] = code.code
                    code = LogCodeCollection.generic
            fields["code"] = code.code
            fields["codeMessage"] = code.code_message
        else:
            if log_level in [LogLevel.ERROR, LogLevel.WARNING]:
                errors.append("Missing LogCode")
                code = LogCodeCollection.generic
                fields["code"] = code.code
                fields["codeMessage"] = code.code_message

        if extra is None:
            fields["extra"] = {}
        elif isinstance(extra, Dict):
            fields["extra"] = extra
        else:
            fields["extra"] = {"data": extra}

        self.log(fields)
        if errors:
            msg = ", ".join(errors)
            traceback_print = ["Traceback (most recent call last):"]
            traceback_print.extend(
                [
                    f'  File "{relpath(frame.filename)}", line {frame.lineno}, in {frame.name}\n'
                    f"    {frame.line}"
                    for frame in traceback.extract_stack()[1:-1]
                ]
            )
            error_fields = {
                "level": LogLevel.ERROR,
                "original": original_log,
                "exception": {
                    "type": "InvalidLogInput",
                    "stack": "\n".join(traceback_print),
                },
                "message": f"Invalid input to logging method.\n{msg}",
                "code": LogCodeCollection.invalid_raw_input_data.code,
                "codeMessage": LogCodeCollection.invalid_raw_input_data.code_message,
            }
            self.log(error_fields)

    def debug(
        self,
        message: str,
        code: Optional[LogCode] = None,
        *,
        extra: Optional[Mapping[str, Any]] = None,
    ):
        """Log a message with DEBUG level"""
        self._log_with_level(message, LogLevel.DEBUG, code, extra=extra)

    def info(
        self,
        message: str,
        code: Optional[LogCode] = None,
        *,
        extra: Optional[Mapping[str, Any]] = None,
    ):
        """Log a message with INFO level"""
        self._log_with_level(message, LogLevel.INFO, code, extra=extra)

    def warning(
        self,
        message: str,
        code: LogCode,
        *,
        extra: Optional[Mapping[str, Any]] = None,
    ):
        """Log a message with WARNING level"""
        self._log_with_level(message, LogLevel.WARNING, code, extra=extra)

    def warn(
        self,
        message: str,
        code: LogCode,
        *,
        extra: Optional[Mapping[str, Any]] = None,
    ):
        """Log a message with WARNING level"""
        self.warning(message, code, extra=extra)

    def error(
        self,
        message: str,
        code: LogCode,
        *,
        extra: Optional[Mapping[str, Any]] = None,
    ):
        """Log a message with ERROR level"""
        self._log_with_level(message, LogLevel.ERROR, code, extra=extra)



import typing as t


def merge_arrays(base="", additions: t.Iterable[str] = []) -> str:
    tags = set(base.split(","))
    for tag in additions:
        if tag in tags:
            raise ValueError(f"Custom tag {tag} is specified multiple times")
        tags.add(tag)
    additions_str = ",".join(additions)
    return _join(base, additions_str, ",")


def merge_objects(base="", additions: t.Mapping[str, str] = {}):
    keys = set(kv.split("=")[0] for kv in base.split("&"))
    for key in additions.keys():
        if key in keys:
            raise ValueError(f"Custom metadata key {key} is already specified")
        keys.add(key)
    additions_str = "&".join(f"{key}={value}" for (key, value) in additions.items())
    return _join(base, additions_str, "&")


def _join(a: str, b: str, sep: str) -> str:
    if a == "":
        return b
    elif b == "":
        return a
    else:
        return a + sep + b



import os
from urllib.parse import urljoin

import requests

from ts_sdk.task.__util_adapters import select_versioned_value, CommunicationFormat


def get(path, params=None, **kwargs):
    return requests.get(get_api_url(path), params=params, **{"verify": False, **kwargs})


def delete(path, **kwargs):
    return requests.delete(get_api_url(path), **{"verify": False, **kwargs})


def patch(path, data=None, **kwargs):
    return requests.patch(get_api_url(path), data=data, **{"verify": False, **kwargs})


def post(path, data=None, json=None, **kwargs):
    return requests.post(
        get_api_url(path), data=data, json=json, **{"verify": False, **kwargs}
    )


def put(path, data=None, **kwargs):
    return requests.put(get_api_url(path), data=data, **{"verify": False, **kwargs})


def get_api_url(path=""):
    try:
        kernel_endpoint = select_versioned_value(
            {
                CommunicationFormat.V0: None,
                CommunicationFormat.V1: os.environ.get("KERNEL_ENDPOINT"),
            }
        )
        if not isinstance(path, str):
            raise ValueError("Path supplied to `get_ts_api_url` must be a string")
        if path != "" and not path.startswith("/"):
            path = "/" + path
        return urljoin(kernel_endpoint, "api") + path
    except NotImplementedError:
        raise NotImplementedError(
            "`get_ts_api_url` is not supported in this environment"
        )



from dataclasses import asdict, is_dataclass

from simplejson import JSONEncoder


class DataclassEncoder(JSONEncoder):
    """Custom json encoder to allow dataclasses to be encoded as JSON objects

    For more about the encoder, see https://docs.python.org/3.7/library/json.html#json.JSONEncode
    """

    def default(self, o):
        """Override the encoding function to allow dataclasses to be encoded.

        For more about dataclass functions, see: https://docs.python.org/3.7/library/dataclasses.html#dataclasses.is_dataclass
        """
        if is_dataclass(o) and not isinstance(o, type):
            return asdict(o)
        return super().default(o)



import json
import os
import re

import boto3
import jsonschema
from ids_validator.instance import validate_ids_instance

from .data_model import IdsUtilDict


def create_ids_util(locations) -> IdsUtilDict:
    def get_location(namespace):
        for location in locations:
            if re.search(location["namespacePattern"], namespace):
                return location
        raise Exception(f"Invalid namespace: {namespace}")

    def get_ids(namespace, slug, version):
        location = get_location(namespace)
        bucket = location["bucket"]
        prefix = location["prefix"]
        endpoint = location["endpoint"]
        if endpoint:
            s3 = boto3.client(
                "s3",
                endpoint_url=endpoint,
                aws_access_key_id="123",
                aws_secret_access_key="abc",
            )
        else:
            s3 = boto3.client("s3")

        file_key = os.path.join(prefix, "ids", namespace, slug, version, "schema.json")
        print(f"downloading ids, {bucket}, {file_key}")
        try:
            read_response = s3.get_object(Bucket=bucket, Key=file_key)
            status_code = read_response.get("ResponseMetadata", {}).get(
                "HTTPStatusCode"
            )
            body = read_response.get("Body").read().decode("utf-8")
            return json.loads(body)
        except Exception as exc:
            raise Exception(
                f"Failed to get IDS, {namespace}/{slug}:{version} from bucket: {bucket}, "
                f"file_key: {file_key}."
            ) from exc

    def validate_ids(data, namespace, slug, version):
        ids = get_ids(namespace, slug, version)
        validate_ids_instance(data, ids)

    return {"get_ids": get_ids, "validate_ids": validate_ids}



import argparse
import io
import os
import re
import zipfile
from pathlib import Path
from time import sleep

import simplejson as json
from ids_validator.ids_validator import validate_ids_using_tdp_artifact
from ids_validator.tdp_api import APIConfig
from loguru import logger

from ts_sdk.cli.__api import TsApi
from ts_sdk.cli.__utils import zipdir
from ts_sdk.cli.put_cmd_helpers.upload_validator import (
    UploadValidator,
    bytes_as_human_readable_string,
)

from ts_sdk.cli.__deprecated import deprecated

DEFAULT_EXCLUDE_FOLDERS = [
    ".git",
    ".venv",
    "example-input",
    "example-output",
    "_tests_",
    "__tests__",
    "__test__",
    "example-files",
    "raw_data",
]


class ArtifactBuildError(Exception):
    """Raise when artifact build process fails"""


def put_cmd_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "type",
        type=str,
        choices=["ids", "protocol", "master-script", "task-script"],
        help="artifact type",
    )
    parser.add_argument("namespace", type=__namespace_type)
    parser.add_argument("slug", type=str.lower)
    parser.add_argument("version", type=__version_type)
    parser.add_argument(
        "folder", type=__folder_type, help="path to folder to be uploaded"
    )
    parser.add_argument(
        "--exclude-folders",
        action="store_true",
        default=True,
        help=f"Whether to exclude the following set of folders: {', '.join(DEFAULT_EXCLUDE_FOLDERS)}",
    )
    parser.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="force overwrite of an existing artifact",
    )
    parser.add_argument(
        "--ignore-ssl",
        "-i",
        action="store_true",
        help="ignore the SSL certificate verification",
    )

    parser.add_argument("--org", help="org slug", type=str)
    parser.add_argument("--api-url", help="platform API URL", type=str)
    parser.add_argument("--auth-token", help="authorization token", type=str)

    parser.add_argument(
        "--config",
        "-c",
        help="JSON file with configuration",
        type=argparse.FileType("r"),
    )

    parser.set_defaults(func=__cmd)


def __version_type(arg_value, pat=re.compile(r"^v")):
    if pat.match(arg_value):
        return arg_value
    return f"v{arg_value}"


def __folder_type(arg_value):
    if os.path.isdir(arg_value):
        return arg_value
    raise argparse.ArgumentTypeError("Not valid folder path provided!")


def __namespace_type(
    arg_value: str, ns_pattern=re.compile("^private-[0-9a-zA-Z-]+$")
) -> str:
    """
    Namespace validation function. Namespace must start with 'private-' and
    only contains alphanumeric characters and hyphens. A hyphen cannot
    succeed another hyphen. An exception is raised for invalid namespaces.

    Args:
        arg_value (str): namespace
        ns_pattern (re.Pattern): namespace validation regex.

     Returns:
         arg_value (str): validated namespace
    """

    # validate namespace
    if (
        not ns_pattern.match(arg_value)
        or arg_value.endswith("-")
        or re.search(r"--", arg_value)
    ):
        raise argparse.ArgumentTypeError(
            f"Invalid namespace {arg_value}. Namespace must start with "
            f"'private-' followed by alphanumeric characters or single "
            f"hyphens."
        )

    return arg_value


def __ensure_args(args: argparse.Namespace):
    # from config
    if getattr(args, "config", None):
        parsed_config = json.load(args.config)
        for k, v in parsed_config.items():
            if getattr(args, k, None) in [None, False]:
                setattr(args, k, v)
    else:
        # from env
        env_prefix = "TS_"
        for k, v in os.environ.items():
            if k.startswith(env_prefix):
                arg_key = k.replace(env_prefix, "").lower()
                if getattr(args, arg_key, None) in [None, False]:
                    setattr(args, arg_key, v)

    args.ignore_ssl = args.ignore_ssl in [True, "true", "True", "1"]


def __validate_ids(args, ts_api: TsApi, force: bool = False) -> None:
    """Run ts-ids-validator on the IDS artifact, raise an exception if it is invalid."""

    print("\n* Validating IDS")
    api_config = APIConfig.from_json_or_env(
        json_config=ts_api.opts,
        json_config_source="ts-sdk config",
    )
    # Validate IDS artifact.
    # API config is used to download the previous IDS for breaking change validation.
    ids_artifact_is_valid = validate_ids_using_tdp_artifact(
        Path(args.folder), api_config=api_config
    )
    if not ids_artifact_is_valid:
        if force:
            print("\n* IDS artifact validation failed, continuing due to 'force' flag.")
        else:
            raise ValueError(
                "IDS artifact validation with ts-ids-validator failed, see the output "
                "of the command for details."
            )


class PutCmdError(Exception):
    """Raised whenever the `ts-sdk put` command encounters an error."""


@deprecated
def __cmd(args):
    __ensure_args(args)
    print("Config:")
    keys_to_show = ["api_url", "org", "auth_token", "ignore_ssl"]
    config_to_show = {
        key_to_show: args.__dict__[key_to_show] for key_to_show in keys_to_show
    }
    if isinstance(config_to_show.get("auth_token"), str):
        config_to_show.update({"auth_token": f'{config_to_show["auth_token"][0:7]}...'})
    print(json.dumps(config_to_show, indent=4, sort_keys=True))

    ts_api = TsApi(**args.__dict__)

    if args.type == "ids":
        # Validate IDS schema.json using ts-ids-validator
        __validate_ids(args, ts_api=ts_api, force=args.force)

    if args.type == "task-script":
        # DE-3436: task-script folder must contain requirements.txt
        package_content = os.listdir(args.folder)
        if "requirements.txt" not in package_content:
            raise Exception("Task-Script package must contain 'requirements.txt'.")

    logger.info("Compressing...", flush=True)
    zip_buffer = io.BytesIO()
    zipf = zipfile.ZipFile(zip_buffer, "a", zipfile.ZIP_DEFLATED, False)
    exclude_folders = []
    if args.exclude_folders:
        exclude_folders = DEFAULT_EXCLUDE_FOLDERS
    zipdir(args.folder, zipf, exclude_folders)
    zipf.close()
    zip_bytes = zip_buffer.getvalue()

    upload_size_checker = UploadValidator()
    maybe_error_messages = upload_size_checker.validate(zip_bytes)
    if len(maybe_error_messages) > 0:
        raise PutCmdError(*maybe_error_messages)

    logger.info(f"Uploading {bytes_as_human_readable_string(len(zip_bytes))}...")

    r = ts_api.upload_artifact(args, zip_bytes)

    build_id = r.get("build", {}).get("id", None)
    if build_id:
        print("Build started", flush=True)
        print(
            "Note: A local script interruption doesn't stop a remote build!", flush=True
        )

        last_status = None
        prev_next_token = ""

        ARTIFACT_BUILD_ERROR_STR = "child process exited with code 1,"

        while True:
            build_info = ts_api.get_task_script_build_info(build_id)
            build_complete = build_info.get("build", {}).get("buildComplete")
            build_status = build_info.get("build", {}).get("buildStatus")

            sleep(3)

            logs_resp = ts_api.get_task_script_build_logs(
                build_id, {"nextToken": prev_next_token}
            )
            prev_next_token = logs_resp.get("nextToken", None)
            events = logs_resp.get("events", [])

            if len(events) > 0:
                print("\r", end="", flush=True)
            elif not build_complete:
                print(".", end="", flush=True)

            for event in events:
                msg_text = event.get("message", "").strip()
                if msg_text:
                    print(msg_text, flush=True)
                    if ARTIFACT_BUILD_ERROR_STR in msg_text:
                        raise ArtifactBuildError(
                            f"Artifact Build process failure:\n{msg_text}"
                        )

            if build_complete:
                last_status = build_status
                break

        print("", flush=True)

        # FAILED: The build failed.
        # FAULT: The build faulted.
        # STOPPED: The build stopped.
        # SUCCEEDED: The build succeeded.
        # TIMED_OUT: The build timed out.

        if last_status == "SUCCEEDED":
            print(
                json.dumps(
                    {
                        "type": args.type,
                        "namespace": args.namespace,
                        "slug": args.slug,
                        "version": args.version,
                    },
                    indent=4,
                    sort_keys=False,
                ),
                flush=True,
            )
        else:
            raise Exception("Build failed.")

    else:
        print(json.dumps(r, indent=4, sort_keys=True), flush=True)



import json
import os
import sys

from main import main

from ts_sdk.taskdev import Context, check_test_output, load_test_file


def test_basic(shared_datadir):
    # set up my for local integration context
    context = Context()
    test_input_file = load_test_file(
        context, shared_datadir / "input.json", "/input.json", file_category="RAW"
    )
    test_input = {
        "input_file": test_input_file,
        "username": "username",
        "password": "secret",
    }

    # do the conversion
    actual_output = main(test_input, context)

    # do the check
    check_test_output(
        context, shared_datadir / "expected.json", actual_output, json.loads
    )



import os
from urllib.parse import urljoin

from ts_sdk.task.__util_adapters import make_adapter, CommunicationFormat
from ts_sdk.task.__util_ts_api import get_api_url


def get_endpoint_adapter():
    return make_adapter(
        {
            CommunicationFormat.V0: {
                "get_public_endpoint": os.environ.get,
            },
            CommunicationFormat.V1: {
                "get_public_endpoint": lambda environment_key: get_api_url(),
            },
        }
    )


def get_public_endpoint(environment_key):
    """
    This submodule is used for accessing endpoints bound in the process's environment
    These endpoints should not be accessed directly, as they may be decorated by this submodule
    """
    return get_endpoint_adapter().get_public_endpoint(environment_key)



import os
from enum import Enum
from itertools import takewhile
from typing import Optional


class CommunicationFormat(Enum):
    V0 = "v0"
    V1 = "v1"
    V2 = "v2"
    VZ = "v?"


DEFAULT_COMMUNICATION_FORMAT = CommunicationFormat.V0
MAXIMUM_COMMUNICATION_FORMAT = CommunicationFormat.V2
UNKNOWN_COMMUNICATION_FORMAT = CommunicationFormat.VZ

COMMUNICATION_FORMAT_ENV_KEY = "PIPELINES_COMMUNICATION_FORMAT"


def get_all_platform_formats():
    """
    :return: An ordered list of CommunicationFormats
    """
    return [communication_format for communication_format in CommunicationFormat]


def to_communication_format(provided_format: Optional[str]) -> CommunicationFormat:
    """
    Coerces a string representation of a communication format into a concreate CommunicationFormat
    :param provided_format: A string representation of a communication format
    :return: CommunicationFormat
    """
    if provided_format is None:
        return DEFAULT_COMMUNICATION_FORMAT
    else:
        search_value = provided_format.lower().strip()
        return next(
            (
                communication_format
                for communication_format in CommunicationFormat
                if communication_format.value == search_value
            ),
            UNKNOWN_COMMUNICATION_FORMAT,
        )


def get_communication_format() -> CommunicationFormat:
    """
    Gets the current CommunicationFormat set in the process's ENV
    :return: CommunicationFormat
    """
    user_communication_format = os.environ.get(COMMUNICATION_FORMAT_ENV_KEY)
    return to_communication_format(user_communication_format)


def get_formats_up_to(communication_format: CommunicationFormat):
    """
    Gets the ordered list of communication formats up to and including the provided CommunicationFormat
    :param communication_format: The final CommunicationFormat to be included in the list of formats
    :return: a list of CommunicationFormats
    """
    preceding_formats = list(
        takewhile(
            lambda other_format: other_format != communication_format,
            get_all_platform_formats(),
        )
    )
    return preceding_formats + [communication_format]
